import re
import sys
import os
import logging
from nomadcore.match_highlighter import ANSI

LOGGER = logging.getLogger(__name__)


# regex for _valid_ fortran float output, what a mess ...
RE_f = (r"(?:" + # outer alternative, between numbers and number-too-wide-for-field markers
    r"([+-]?)(?:" + # MANTISSA, SIGN (group 1, optional), followed by alternatives
        '|'.join([ # MANTISSA
            r"(\d+(?!\.))", # MANTISSA without a decimal point, group 2
            r"(\d*)" + ( # MANTISSA, WHOLE part (group 3)
                # we need negative look-ahead/look-behind assertions around the
                # decimal point as there is too much optional stuff around
                r"(?<![^\d\s+-])" + # char preceding the dot must be nothing but number, whitespace, or sign
                r"\." +
                r"(?![^eEdD\d\s,])" + # char succeeding the dot must be nothing but number, exponential/precision char, comma or whitespace
                r"(\d*)" # MANTISSA, FRACTIONAL part (group 4), separated by dot
            )
        ]) +
    r")(?:" + ( # EXPONENT part (optional)
        r"([eEdD])" + # PRECISION (group5)
        r"([+-]?)(\d*)" # EXPONENT SIGN (group 6), VALUE (group 7)
    ) + ")?" + # make precision/exponet part optinal
    r"|(\*+))" # outer alternative, between numbers and number-too-wide markers (group 8)
)
cRE_f = re.compile(RE_f)


def match_to_float(m, group_offset=0):
    group = [ m.group(0) ] + [ m.group(group_offset + i) for i in range(1,9)]
    LOGGER.debug("g: %s", str(group))
    if group[8] is not None:
        pyfloat_str = 'nan'
        dtype = 'f'
    else:
        pyfloat_str = group[1] # sign, maybe zero-length
        if group[2] is not None:
            pyfloat_str += group[2]
            dtype = 'i'
        else:
            pyfloat_str += group[3] if len(group[3])>0 else '0'
            pyfloat_str += '.'
            pyfloat_str += group[4] if len(group[4])>0 else '0'
            dtype = 'f'
        if group[5] is not None:
            pyfloat_str += 'e' + group[6]
            pyfloat_str += group[7] if len(group[7])>0 else '0'
            dtype = 'f'
    LOGGER.debug("pyfloat_str: %s", pyfloat_str)
    if dtype == 'f':
        return (float(pyfloat_str), dtype)
    else:
        return (int(pyfloat_str), dtype)

RE_unescape = {
    '"': re.compile(r'""'),
    "'": re.compile(r"''"),
}


def unquote_string(value):
    result = value[1:-1]
    result = RE_unescape[value[0]].sub(value[0], result)
    return result


# quoted strings
cRE_string_quoted = re.compile(r"(?:'[^']*'|\"[^\"]*\")")
cRE_comment = re.compile(r"\s*!(?P<comment>.*)")
RE_identifier = r"[a-zA-Z]\w*" # fortran identifier
cRE_start_group = re.compile(r'\s*&(' + RE_identifier + r')') # beginning of namelist group
cRE_end_group = re.compile(r'\s*/')
cRE_assigned_value = re.compile(
    r'\s*(?:' + '|'.join([
        r'(?P<num>' + RE_f + r')', # integers and floats
        r'\(\s*(?P<cnum_r>' + RE_f + r')\s*,\s*(?P<cnum_i>' + RE_f + r')\s*\)', # complex numbers
        r'(?P<bool_t>\.t(?:rue)?\.)', # true-value bool
        r'(?P<bool_f>\.f(?:alse)?\.)', # false-value bool
        r"(?P<str_s>'[^']*(?:[^']|'')*'(?!'))", # single-quoted string, closed, allowing for escaped quotes ('')
        r'(?P<str_d>"[^"]*(?:[^"]|"")*"(?!"))', # double-quoted string, closed, allowing for escaped quotes ("")
        r"(?P<str_s_nc>'[^']*(?:[^']|'')*)", # single-quoted string, not closed
        r'(?P<str_d_nc>"[^"]*(?:[^"]|"")*)', # double-quoted string, not closed
        r'!(?P<comment>.*)', # comment
    ]) + ')', re.I)
cRE_str_s_close = re.compile(r"([^']*(?:[^']|'')*'(?!'))") # single-quoted string, closing
cRE_str_d_close = re.compile(r'([^"]*(?:[^"]|"")*"(?!"))') # double-quoted string, closing
cRE_comma = re.compile(r'\s*,')
cRE_trailing_whitespace = re.compile(r'\s+$')

cRE_identifier = re.compile(r'\s*(?P<target>' + RE_identifier + r')')
cRE_assignment_subscript_open = re.compile(r'\s*\((?P<subscript>[^\)!]*)')
cRE_assignment_subscript_continue = re.compile(r'(?P<subscript>[^\)!]+)')
cRE_assignment_subscript_close = re.compile(r'(?P<subscript>[^\)!]*)\)')
cRE_assignment_equals = re.compile(r'\s*=')

cRE_subscript = re.compile(r'\s*,?\s*(?:(\d*)\s*:\s*(\d*)|(\d+))')

cRE_end_newline = re.compile(r'(.*?)(\n*)$')

class FortranNamelistParser(object):
    """Parser for Fortran 90 Namelists
    """
    def __init__(self, file_path, annotateFile = None):
        self.input_tree = {}
        self.file_path = file_path
        self.state = self.state_root
        self.__annotateFile = annotateFile
        self.__nl_group = None
        self.__target = None
        self.__subscript = None
        self.__values = None
        self.__types = None
        self.__nvalues_after_comma = 0
        self.__cre_closing = None
        self.bad_input = False
        self.cache = {}

    def parse(self):
        """open file and parse line-by-line"""
        with open(self.file_path, "r") as fIn:
            # process line-by-line
            for line in fIn:
                self.parse_line(line)
        # check if there was input flagged as 'bad'/'syntactically incorrect'
        if self.bad_input:
            # call bad-input hook
            self.onBad_input()
        # call end-of-file hook
        self.onEnd_of_file()

    def parse_line(self, line):
        """parse one line, delegating to the parser state handlers"""
        pos_in_line = 0
        while pos_in_line<len(line):
            new_pos_in_line = self.state(line, pos_in_line)
            # check if anything was parsed, otherwise cancel that line
            if new_pos_in_line is None:
                break
            else:
                pos_in_line = new_pos_in_line
        if pos_in_line < len(line):
            self.bad_input = True
            self.annotate(line[pos_in_line:], ANSI.BEGIN_INVERT + ANSI.FG_BRIGHT_RED)

    def annotate(self, what, highlight):
        """write string to annotateFile with ANSI highlight/reset sequences"""
        if self.__annotateFile:
            m = cRE_end_newline.match(what)
            self.__annotateFile.write(highlight + m.group(1) + ANSI.RESET + m.group(2))

    def parse_subscript_string(self, subscript):
        """parse fully captured subscript string into python array"""
        if subscript is None:
            return None
        result = []
        last_end = 0
        while last_end<len(subscript):
            m = cRE_subscript.match(subscript, last_end)
            if m is None:
                break
            elif m.group(3) is not None:
                # prepend to result list, making ranges explicit:
                #    fortran has fastest-running index first
                #  while
                #    python/c has fastest-running index last
                result[0:0] = [ [int(m.group(3))] ]
                last_end = m.end()
                continue
            elif m.group(1) is not None:
                # prepend to result list, making ranges explicit
                #    fortran has fastest-running index first
                #  while
                #    python/c has fastest-running index last
                result[0:0] = [list(range(int(m.group(1)),int(m.group(2))+1))]
                last_end = m.end()
                continue
            break
        if last_end < len(subscript):
            if subscript[last_end:].strip():
                LOGGER.error("ERROR: leftover chars in subscript: '%s'", subscript[last_end:])
                self.bad_input = True
        return result

    def state_root(self, line, pos_in_line):
        """state: no open namelist groups, i.e. at the root of the namelist"""
        m = cRE_start_group.match(line, pos_in_line)
        if m is not None:
            self.__nl_group = m.group(1).lower()
            self.annotate(m.group(), ANSI.FG_BRIGHT_GREEN)
            self.state = self.state_inside_group
            self.onOpen_namelist_group(self.__nl_group)
            return m.end()
        else:
            # but comments may appear here
            m = cRE_comment.match(line, pos_in_line)
            if m is not None:
                self.annotate(m.group(), ANSI.FG_BLUE)
                self.onComment(m.group('comment'))
                return m.end()
            # as well as whitespace-only lines
            m = cRE_trailing_whitespace.match(line, pos_in_line)
            if m is not None:
                self.annotate(m.group(), ANSI.BG_WHITE)
                return m.end()
        # nothing matched, call hook
        return self.onRoot_data(line, pos_in_line)

    def state_inside_group(self, line, pos_in_line):
        """state: inside opened group, but no open assignment"""
        # check for group-closing /
        m = cRE_end_group.match(line, pos_in_line)
        if m is not None:
            # we just closed a NL group
            self.annotate(m.group(), ANSI.BEGIN_INVERT + ANSI.FG_BRIGHT_GREEN)
            if self.__target is not None:
                self.onClose_value_assignment(
                    self.__nl_group,
                    self.__target, self.__subscript,
                    self.__values, self.__types)
            self.__target = None
            self.__subscript = None
            self.__values = None
            self.__types = None
            self.__nvalues_after_comma = 0
            self.onClose_namelist_group(self.__nl_group)
            self.__nl_group = None
            self.state = self.state_root
            return m.end()
        # check for new identifier (part of left-hand side of assignment)
        m = cRE_identifier.match(line, pos_in_line)
        if m is not None:
            self.annotate(m.group(), ANSI.FG_GREEN)
            if self.__target is not None:
                self.onClose_value_assignment(
                    self.__nl_group,
                    self.__target, self.__subscript,
                    self.__values, self.__types)
            self.__target = m.group('target').lower()
            self.__subscript = None
            self.__values = []
            self.__types = []
            self.__nvalues_after_comma = 0
            return m.end()
        # check for new subscript (part of left-hand side of assignment)
        m = cRE_assignment_subscript_open.match(line, pos_in_line)
        if m is not None:
            self.annotate(line[pos_in_line:m.start('subscript')], ANSI.FG_GREEN)
            self.annotate(m.group('subscript'), ANSI.FG_CYAN)
            self.__subscript = m.group('subscript')
            self.state = self.state_assignment_subscript
            return m.end()
        # check for '=' sign in assignment, separating target from values
        m = cRE_assignment_equals.match(line, pos_in_line)
        if m is not None:
            self.annotate(line[pos_in_line:m.end()], ANSI.FG_GREEN)
            self.onOpen_value_assignment(
                self.__nl_group,
                self.__target, self.__subscript)
            self.state = self.state_assignment_values
            return m.end()
        # check for comments ('!' character up until end of line)
        m = cRE_comment.match(line, pos_in_line)
        if m is not None:
            self.annotate(m.group(), ANSI.FG_BLUE)
            self.onComment(m.group('comment'))
            return m.end()
        # check for trailing whitespace
        m = cRE_trailing_whitespace.match(line, pos_in_line)
        if m is not None:
            self.annotate(m.group(), ANSI.BG_WHITE)
            return m.end()
        return None

    def state_assignment_values(self, line, pos_in_line):
        """state: parse values, i.e. right-hand side of assignment"""
        # match value literals, groups decide on data type
        m = cRE_assigned_value.match(line, pos_in_line)
        if m is not None:
            if m.group('comment') is not None:
                # found a comment
                self.annotate(m.group(), ANSI.FG_BLUE)
                self.onComment(m.group('comment'))
            else:
                self.annotate(m.group(), ANSI.FG_YELLOW)
                if m.group('num') is not None:
                    # literal is single integer or float
                    (value, dtype) = match_to_float(m, group_offset=1)
                    self.__values.append(value)
                    self.__types.append(dtype)
                elif m.group('cnum_r') is not None:
                    # literal is complex: (float, float)
                    (cnum_r, dtype) = match_to_float(m, group_offset=10)
                    (cnum_i, dtype) = match_to_float(m, group_offset=19)
                    self.__values.append(complex(cnum_r, cnum_i))
                    self.__types.append('complex')
                elif m.group('bool_t') is not None:
                    # literal is a true-value bool
                    self.__values.append(True)
                    self.__types.append('b')
                elif m.group('bool_f') is not None:
                    # literal is a false-value bool
                    self.__values.append(False)
                    self.__types.append('b')
                elif m.group('str_s') is not None:
                    # literal is a closed, single-quoted string
                    self.__values.append(unquote_string(m.group('str_s')))
                    self.__types.append('C')
                elif m.group('str_d') is not None:
                    # literal is a closed, double-quoted string
                    self.__values.append(unquote_string(m.group('str_d')))
                    self.__types.append('C')
                elif m.group('str_s_nc') is not None:
                    # literal is a non-closed, single-quoted string
                    self.state = self.state_assignment_values_multiline_string
                    self.__values.append(m.group('str_s_nc'))
                    self.__types.append('C')
                    self.__cre_closing = cRE_str_s_close
                elif m.group('str_d_nc') is not None:
                    # literal is a non-closed, double-quoted string
                    self.state = self.state_assignment_values_multiline_string
                    self.__values.append(m.group('str_d_nc'))
                    self.__types.append('C')
                    self.__cre_closing = cRE_str_d_close
                # keep track if there were values following the previous comma
                self.__nvalues_after_comma += 1
            return m.end()
        # special meaning of comma: may indicate Null values in array assignments
        m = cRE_comma.match(line, pos_in_line)
        if m is not None:
            self.annotate(m.group(), ANSI.FG_MAGENTA)
            if self.__nvalues_after_comma is 0:
                # there were no value literals preceeding the comma, which
                # means a null value
                self.__values.append(None)
                self.__types.append(None)
            self.__nvalues_after_comma = 0
            return m.end()
        # check for trailing whitespace
        m = cRE_trailing_whitespace.match(line, pos_in_line)
        if m is not None:
            self.annotate(m.group(), ANSI.BG_WHITE)
            return m.end()
        # if none of the above matched, switch back to checking for new assignment
        self.state = self.state_inside_group
        return pos_in_line

    def state_assignment_values_multiline_string(self, line, pos_in_line):
        """state: parse multiline string in right-hand side of assignment"""
        # check for closing quotes
        m = self.__cre_closing.match(line, pos_in_line)
        if m is None:
            # no closing quotes, append data to value
            self.annotate(line[pos_in_line:], ANSI.FG_YELLOW)
            self.__values[-1] += line
            return len(line)
        else:
            # closing quotes, postprocess string
            self.annotate(m.group(), ANSI.FG_YELLOW)
            self.__values[-1] += m.group(1)
            # remove enclosing quotes and resolve escaped quotes in string
            self.__values[-1] = unquote_string(self.__values[-1])
            self.__cre_closing = None
            self.state = self.state_assignment_values
            return m.end()
        return None

    def state_assignment_subscript(self, line, pos_in_line):
        """state: capture subscipt, possibly spanning multiple lines"""
        # check for closing bracket
        m = cRE_assignment_subscript_close.match(line, pos_in_line)
        if m is not None:
            # subscript closed, convert string form to python array
            self.annotate(m.group('subscript'), ANSI.FG_CYAN)
            self.annotate(line[m.end('subscript'):m.end()], ANSI.FG_GREEN)
            self.__subscript = self.parse_subscript_string(self.__subscript + m.group('subscript'))
            self.state = self.state_inside_group
            return m.end()
        # check for new indices in subscript
        m = cRE_assignment_subscript_continue.match(line, pos_in_line)
        if m is not None:
            self.annotate(m.group('subscript'), ANSI.FG_CYAN)
            self.__subscript += m.group('subscript')
            return m.end()
        # comments may appear within subscripts spanning multiple lines
        m = cRE_comment.match(line, pos_in_line)
        if m is not None:
            self.annotate(m.group(), ANSI.FG_BLUE)
            self.onComment(m.group('comment'))
            self.__subscript += line[pos_in_line:m.start()]
            return m.end()
        self.annotate(line[pos_in_line:], ANSI.BEGIN_INVERT + ANSI.FG_BRIGHT_RED)
        LOGGER.error("ERROR: leftover chars in line while inside subscript: '%s'", line[pos_in_line:])
        self.bad_input = True
        return None

    # Hooks to be overloaded in derived classes in order to do stuff beyond caching
    def onComment(self, comment):
        """hook: called whan a comment was found"""
        pass

    def onOpen_namelist_group(self, groupname):
        """hook: called when a namelist group opens"""
        if groupname in self.cache:
            LOGGER.error("ERROR: multiple definitions of group &%s", groupname)
            self.bad_input = True
        else:
            self.cache[groupname]={}

    def onClose_namelist_group(self, groupname):
        """hook: called when a namelist group closes"""
        LOGGER.error("group: %s", groupname)
        for identifier in sorted(self.cache[groupname]):
            LOGGER.error("  %s: %s", identifier, str(self.cache[groupname][identifier]))

    def onOpen_value_assignment(self, groupname, target, subscript):
        """hook: called when a value assignment within a namelist group starts"""
        pass

    def onClose_value_assignment(self, groupname, target, subscript, values, dtypes):
        """hook: called when a value assignment within a namelist group closes
        Arguments are: NL group name, identifier/subscript, values and assumed
        data types
        """
        if subscript is None:
            LOGGER.debug("SET %s/%s = %s (types: %s)", groupname, target, str(values), str(dtypes))
        else:
            LOGGER.debug("SET %s/%s(%s) = %s (types: %s)", groupname, target, subscript, str(values), str(dtypes))
        if target not in self.cache[groupname]:
            self.cache[groupname][target] = []
        self.cache[groupname][target].append([subscript, values, dtypes])

    def onRoot_data(self, line, pos_in_line):
        """hook: called if data appears outside namelists groups, directly
        at root level within the file;
        data means: line is not empty or a comment
        useful for code-specific extensions beyond the F90 namelist standard
        """
        return None

    def onBad_input(self):
        """hook: called at the end of parsing if there was any bad input"""
        pass

    def onEnd_of_file(self):
        """hook: called at the end of parsing"""
        pass

if __name__ == "__main__":
    parser = FortranNamelistParser(sys.argv[1], annotateFile=sys.stdout)
    parser.parse()



""" OPTIMADE landing page, rendered as a Jinja2 template. """

from pathlib import Path
from fastapi.templating import Jinja2Templates
from starlette.routing import Router, Route
from optimade import __api_version__

from . import ENTRY_COLLECTIONS
from .utils import meta_values

template_dir = Path(__file__).parent.joinpath("static").resolve()
TEMPLATES = Jinja2Templates(directory=[template_dir])


async def landing(request):
    """ Show a human-readable landing page when the base URL is accessed. """

    meta = meta_values(str(request.url), 1, 1, more_data_available=False)

    major_version = __api_version__.split(".")[0]
    versioned_url = (
        f"{request.url}"
        if f"v{major_version}" in f"{request.url.path}"
        else f"{request.url}v{major_version}/"
    )

    context = {
        "request": request,
        "request_url": request.url,
        "api_version": __api_version__,
        "implementation": meta.implementation,
        "versioned_url": versioned_url,
        "provider": meta.provider,
        "endpoints": list(ENTRY_COLLECTIONS.keys()) + ["info"],
    }

    return TEMPLATES.TemplateResponse("landing_page.html", context)


router = Router(routes=[Route("/", endpoint=landing)])



from builtins import object
from nomadcore.simple_parser import mainFunction, CachingLevel
from nomadcore.simple_parser import SimpleMatcher as SM
from nomadcore.local_meta_info import loadJsonFile, InfoKindEl
import os, sys, json


################################################################
# This is the subparser for the WIEN2k input file (.in2c)
################################################################

# Copyright 2016-2018 Daria M. Tomecka, Fawzi Mohamed
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = "Daria M. Tomecka"
__maintainer__ = "Daria M. Tomecka"
__email__ = "tomeckadm@gmail.com;"
__date__ = "15/05/2017"


class Wien2kIn2cContext(object):
    """context for wien2k In2c parser"""

    def __init__(self):
        self.parser = None

    def initialize_values(self):
        """allows to reset values if the same superContext is used to parse different files"""
        pass

    def startedParsing(self, path, parser):
        """called when parsing starts"""
        self.parser = parser
        # allows to reset values if the same superContext is used to parse different files
        self.initialize_values()


# description of the input
def buildIn2cMatchers():
    return SM(
    name = 'root',
    weak = True,
    startReStr = "",
        sections = ["section_run", "section_method"],
    subMatchers = [
        SM(r"\s*(?P<x_wien2k_in2c_switch>[A-Z]+)\s*.*"),
        SM(r"\s*(?P<x_wien2k_in2c_emin>[-+0-9.]+)\s*(?P<x_wien2k_in2c_ne>[-+0-9.]+)\s*(?P<x_wien2k_in2c_espermin>[-+0-9.]+)\s*(?P<x_wien2k_in2c_esper0>[-+0-9.]+)\s*.*"),
        SM(r"\s*(?P<x_wien2k_smearing_kind>[A-Z]+)\s*\s*(?P<x_wien2k_smearing_width__rydberg>[-+0-9.]+)\s*.*"),
        SM(r"\s*(?P<x_wien2k_in2c_gmax>[-+0-9.]+)\s*GMAX")
    ])

def get_cachingLevelForMetaName(metaInfoEnv, CachingLvl):
    """Sets the caching level for the metadata.

    Args:
        metaInfoEnv: metadata which is an object of the class InfoKindEnv in nomadcore.local_meta_info.py.
        CachingLvl: Sets the CachingLevel for the sections k_band, run, and single_configuration_calculation.
            This allows to run the parser without opening new sections.

    Returns:
        Dictionary with metaname as key and caching level as value.
    """
    # manually adjust caching of metadata
    cachingLevelForMetaName = {
                               'section_run': CachingLvl,
                               'section_method': CachingLvl
                              }
    return cachingLevelForMetaName

# loading metadata from nomad-meta-info/meta_info/nomad_meta_info/fhi_aims.nomadmetainfo.json



# Copyright 2015-2018 Lauri Himanen, Fawzi Mohamed, Ankit Kariryaa
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from __future__ import print_function
from __future__ import absolute_import
from builtins import next
from builtins import range
from nomadcore.simple_parser import SimpleMatcher as SM
from nomadcore.baseclasses import MainHierarchicalParser
import nomadcore.configurationreading
import nomadcore.csvparsing
from .commonparser import CP2KCommonParser
from nomadcore.caching_backend import CachingLevel
import logging
logger = logging.getLogger("nomad")


class CP2KGeoOptParser(MainHierarchicalParser):
    """Used to parse the CP2K calculation with run types:
        -GEO_OPT/GEOMETRY_OPTIMIZATION
    """
    def __init__(self, parser_context):
        """
        """
        super(CP2KGeoOptParser, self).__init__(parser_context)
        self.setup_common_matcher(CP2KCommonParser(parser_context))
        self.traj_iterator = None
        self.energy_reeval_quickstep = None

        #=======================================================================
        # Globally cached values
        self.cache_service.add("number_of_frames_in_sequence", 0)
        self.cache_service.add("frame_sequence_potential_energy", [])
        self.cache_service.add("frame_sequence_local_frames_ref", [])
        self.cache_service.add("geometry_optimization_method")

        #=======================================================================
        # Cache levels
        self.caching_levels.update({
            'x_cp2k_section_geometry_optimization_step': CachingLevel.ForwardAndCache,
            'x_cp2k_section_quickstep_calculation': CachingLevel.ForwardAndCache,
            'x_cp2k_section_geometry_optimization': CachingLevel.ForwardAndCache,
            # 'x_cp2k_section_geometry_optimization_energy_reevaluation': CachingLevel.ForwardAndCache,
        })

        #=======================================================================
        # SimpleMatchers
        self.geo_opt = SM(
            " ***                     STARTING GEOMETRY OPTIMIZATION                      ***".replace("*", "\*"),
            sections=["section_frame_sequence", "x_cp2k_section_geometry_optimization"],
            subMatchers=[
                SM( " ***                           CONJUGATE GRADIENTS                           ***".replace("*", "\*"),
                    adHoc=self.adHoc_conjugate_gradient(),
                    otherMetaInfo=["geometry_optimization_method"],
                ),
                SM( " ***                                   BFGS                                  ***".replace("*", "\*"),
                    adHoc=self.adHoc_bfgs(),
                    otherMetaInfo=["geometry_optimization_method"],
                ),
                SM( " ***                                 L-BFGS                                  ***".replace("*", "\*"),
                    adHoc=self.adHoc_bfgs(),
                    otherMetaInfo=["geometry_optimization_method"],
                ),
                # SM( "",
                    # forwardMatch=True,
                    # sections=["section_single_configuration_calculation", "section_system", "x_cp2k_section_geometry_optimization_step"],
                    # subMatchers=[
                        # self.cm.quickstep_calculation(),
                        # SM( " --------  Informations at step"),
                        # SM( "  Optimization Method        =\s+(?P<x_cp2k_optimization_method>{})".format(self.regexs.word)),
                        # SM( "  Total Energy               =\s+(?P<x_cp2k_optimization_energy__hartree>{})".format(self.regexs.float),
                            # otherMetaInfo=["frame_sequence_potential_energy"]
                        # ),
                    # ],
                    # otherMetaInfo=["atom_positions"],
                    # adHoc=self.adHoc_step(),
                # ),
                SM( " OPTIMIZATION STEP:",
                    endReStr="  Conv. in RMS gradients     =",
                    name="geooptstep",
                    repeats=True,
                    subMatchers=[
                        SM( "",
                            forwardMatch=True,
                            sections=["x_cp2k_section_geometry_optimization_step"],
                            otherMetaInfo=[
                                "atom_positions",
                            ],
                            subMatchers=[
                                # SM( "",
                                    # forwardMatch=True,
                                    # endReStr=" ***                 MNBRACK - NUMBER OF ENERGY EVALUATIONS :\s+{}\s+***".replace("*", "\*").format(self.regexs.int),
                                    # subMatchers=[
                                        # SM(" SCF WAVEFUNCTION OPTIMIZATION",
                                            # forwardMatch=True,
                                            # repeats=True,
                                            # subMatchers=[
                                                # self.cm.quickstep_calculation(),
                                            # ]
                                        # )
                                    # ]
                                # ),
                                # SM( "",
                                    # forwardMatch=True,
                                    # endReStr=" ***                 BRENT   - NUMBER OF ENERGY EVALUATIONS :\s+{}\s+***".replace("*", "\*").format(self.regexs.int),
                                    # subMatchers=[
                                        # SM(" SCF WAVEFUNCTION OPTIMIZATION",
                                            # forwardMatch=True,
                                            # repeats=True,
                                            # subMatchers=[
                                                # self.cm.quickstep_calculation(),
                                            # ]
                                        # )
                                    # ]
                                # ),
                                SM( " --------  Informations at step"),
                                SM( "  Optimization Method        =\s+(?P<x_cp2k_optimization_method>{})".format(self.regexs.word)),
                                SM( "  Total Energy               =\s+(?P<x_cp2k_optimization_energy__hartree>{})".format(self.regexs.float),
                                    otherMetaInfo=["frame_sequence_potential_energy"]
                                ),
                                SM( "  Real energy change         =\s+(?P<x_cp2k_optimization_energy_change__hartree>{})".format(self.regexs.float)),
                                SM( "  Decrease in energy         =\s+(?P<x_cp2k_optimization_energy_decrease>{})".format(self.regexs.word)),
                                SM( "  Used time                  =\s+(?P<x_cp2k_optimization_used_time>{})".format(self.regexs.float)),
                                SM( "  Max. step size             =\s+(?P<x_cp2k_optimization_max_step_size__bohr>{})".format(self.regexs.float)),
                                SM( "  Conv. limit for step size  =\s+(?P<x_cp2k_optimization_step_size_convergence_limit__bohr>{})".format(self.regexs.float),
                                    otherMetaInfo=["geometry_optimization_geometry_change"]
                                ),
                                SM( "  Convergence in step size   =\s+(?P<x_cp2k_optimization_step_size_convergence>{})".format(self.regexs.word)),
                                SM( "  RMS step size              =\s+(?P<x_cp2k_optimization_rms_step_size__bohr>{})".format(self.regexs.float)),
                                SM( "  Convergence in RMS step    =\s+(?P<x_cp2k_optimization_rms_step_size_convergence>{})".format(self.regexs.word)),
                                SM( "  Max. gradient              =\s+(?P<x_cp2k_optimization_max_gradient__bohr_1hartree>{})".format(self.regexs.float)),
                                SM( "  Conv. limit for gradients  =\s+(?P<x_cp2k_optimization_gradient_convergence_limit__bohr_1hartree>{})".format(self.regexs.float),
                                    otherMetaInfo=["geometry_optimization_threshold_force"]
                                ),
                                SM( "  Conv. for gradients        =\s+(?P<x_cp2k_optimization_max_gradient_convergence>{})".format(self.regexs.word)),
                                SM( "  RMS gradient               =\s+(?P<x_cp2k_optimization_rms_gradient__bohr_1hartree>{})".format(self.regexs.float)),
                                SM( "  Conv. in RMS gradients     =\s+(?P<x_cp2k_optimization_rms_gradient_convergence>{})".format(self.regexs.word)),
                            ],
                            # adHoc=self.adHoc_step()
                        ),
                    ]
                ),
                SM( " ***                    GEOMETRY OPTIMIZATION COMPLETED                      ***".replace("*", "\*"),
                    adHoc=self.adHoc_geo_opt_converged(),
                    otherMetaInfo=["geometry_optimization_converged"]
                ),
                SM( "                    Reevaluating energy at the minimum",
                    # sections=["x_cp2k_section_geometry_optimization_energy_reevaluation"],
                    subMatchers=[
                        self.cm.quickstep_calculation(),
                        # SM("",
                            # adHoc=self.adHoc_save_energy_reeval_quickstep()
                        # )
                    ],
                    # adHoc=self.adHoc_save_energy_reeval_quickstep()
                ),
                # SM( "",
                    # forwardMatch=True,
                    # adHoc=self.adHoc_save_energy_reeval_quickstep()
                # )
            ],
        )

        # Compose root matcher according to the run type. This way the
        # unnecessary regex parsers will not be compiled and searched. Saves
        # computational time.
        self.root_matcher = SM("",
            forwardMatch=True,
            sections=["section_run", "section_sampling_method"],
            subMatchers=[
                SM( "",
                    forwardMatch=True,
                    sections=["section_method"],
                    subMatchers=[
                        self.cm.header(),
                        self.cm.quickstep_header(),
                    ],
                ),
                self.geo_opt,
                self.cm.footer(),
            ]
        )

    #===========================================================================
    # onClose triggers
    def onClose_x_cp2k_section_geometry_optimization(self, backend, gIndex, section):

        # Get the re-evaluated energy and add it to frame_sequence_potential_energy
        reeval_quickstep = self.energy_reeval_quickstep
        if reeval_quickstep is not None:
            energy = reeval_quickstep.get_latest_value("x_cp2k_energy_total")
            if energy is not None:
                self.cache_service["frame_sequence_potential_energy"].append(energy)

        # Push values from cache
        self.cache_service.addArrayValues("frame_sequence_potential_energy")
        self.cache_service.addValue("geometry_optimization_method")
        self.backend.addValue("frame_sequence_to_sampling_ref", 0)

        # Get the optimization convergence criteria from the last optimization
        # step
        section.add_latest_value([
            "x_cp2k_section_geometry_optimization_step",
            "x_cp2k_optimization_step_size_convergence_limit"],
            "geometry_optimization_geometry_change",
        )
        section.add_latest_value([
            "x_cp2k_section_geometry_optimization_step",
            "x_cp2k_optimization_gradient_convergence_limit"],
            "geometry_optimization_threshold_force",
        )

        # Push the information into single configuration and system
        steps = section["x_cp2k_section_geometry_optimization_step"]
        each = self.cache_service["each_geo_opt"]
        add_last = False
        add_last_setting = self.cache_service["traj_add_last"]
        if add_last_setting == "NUMERIC" or add_last_setting == "SYMBOLIC":
            add_last = True

        # Push the trajectory
        n_steps = len(steps) + 1
        last_step = n_steps - 1

        # First push the original system geometry
        # print(self.cache_service["atom_positions"])
        for i_step in range(n_steps):
            singleId = backend.openSection("section_single_configuration_calculation")
            systemId = backend.openSection("section_system")

            if self.traj_iterator is not None:
                if (i_step + 1) % each == 0 or (i_step == last_step and add_last):
                    try:
                        pos = next(self.traj_iterator)
                    except StopIteration:
                        logger.error(
                            "Could not get next geometry from an external"
                            " file. It seems that the number of optimization "
                            "steps in the CP2K output file doesn't match the "
                            "number of steps found in the external trajectory "
                            "file."
                        )
                    else:
                        backend.addArrayValues("atom_positions", pos, unit="angstrom")
            backend.closeSection("section_system", systemId)
            backend.closeSection("section_single_configuration_calculation", singleId)

        self.cache_service.addArrayValues("frame_sequence_local_frames_ref")
        backend.addValue("number_of_frames_in_sequence", n_steps)

    def onClose_section_sampling_method(self, backend, gIndex, section):
        self.backend.addValue("sampling_method", "geometry_optimization")

    def onClose_x_cp2k_section_quickstep_calculation(self, backend, gIndex, section):
        self.energy_reeval_quickstep = section

    def onClose_x_cp2k_section_geometry_optimization_step(self, backend, gIndex, section):
        energy = section["x_cp2k_optimization_energy"]
        if energy is not None:
            self.cache_service["frame_sequence_potential_energy"].append(energy[0])

    def onClose_section_system(self, backend, gIndex, section):
        self.cache_service.addArrayValues("simulation_cell", unit="angstrom")
        self.cache_service.addArrayValues("lattice_vectors", unit="angstrom")

    def onClose_section_method(self, backend, gIndex, section):
        traj_file = self.file_service.get_file_by_id("trajectory")
        traj_format = self.cache_service["trajectory_format"]
        if traj_format is not None and traj_file is not None:

            # Use special parsing for CP2K pdb files because they don't follow the proper syntax
            if traj_format == "PDB":
                self.traj_iterator = nomadcore.csvparsing.iread(traj_file, columns=[3, 4, 5], start="CRYST", end="END")
            else:
                try:
                    self.traj_iterator = nomadcore.configurationreading.iread(traj_file)
                except ValueError:
                    pass

    def onClose_section_single_configuration_calculation(self, backend, gIndex, section):
        self.cache_service["frame_sequence_local_frames_ref"].append(gIndex)

    #===========================================================================
    # adHoc functions
    def adHoc_geo_opt_converged(self):
        """Called when the geometry optimization converged.
        """
        def wrapper(parser):
            parser.backend.addValue("geometry_optimization_converged", True)
        return wrapper

    def adHoc_geo_opt_not_converged(self):
        """Called when the geometry optimization did not converge.
        """
        def wrapper(parser):
            parser.backend.addValue("geometry_optimization_converged", False)
        return wrapper

    def adHoc_conjugate_gradient(self):
        """Called when conjugate gradient method is used.
        """
        def wrapper(parser):
            self.cache_service["geometry_optimization_method"] = "conjugate_gradient"
        return wrapper

    def adHoc_bfgs(self):
        """Called when conjugate gradient method is used.
        """
        def wrapper(parser):
            self.cache_service["geometry_optimization_method"] = "bfgs"
        return wrapper

    # def adHoc_save_energy_reeval_quickstep(self):
        # def wrapper(parser):
            # section_managers = parser.backend.sectionManagers
            # section_run_manager = section_managers["section_run"]
            # section_run = section_run_manager.openSections[0]
            # print section_run.subSectionValues
            # # quickstep = section_run.get_latest_value("x_cp2k_section_quickstep_calculation")
            # # print quickstep
            # # self.energy_reeval_quickstep = quickstep
        # return wrapper

    def debug(self):
        def wrapper(parser):
            print("DEBUG")
        return wrapper



# pylint: disable=no-self-argument,line-too-long,no-name-in-module
from enum import IntEnum
from sys import float_info
from typing import List, Optional, Tuple, Union

from pydantic import Field, BaseModel, validator

from .entries import EntryResourceAttributes, EntryResource
from .utils import CHEMICAL_SYMBOLS, EXTRA_SYMBOLS


EXTENDED_CHEMICAL_SYMBOLS = CHEMICAL_SYMBOLS + EXTRA_SYMBOLS


__all__ = ("Species", "Assembly", "StructureResourceAttributes", "StructureResource")


EPS = float_info.epsilon


Vector3D = Tuple[Union[float, None], Union[float, None], Union[float, None]]


class Periodicity(IntEnum):
    APERIODIC = 0
    PERIODIC = 1


class Species(BaseModel):
    """A list describing the species of the sites of this structure.
    Species can be pure chemical elements, or virtual-crystal atoms representing a statistical occupation of a given site by multiple chemical elements.

    - **Examples**:

        - :val:`[ {"name": "Ti", "chemical_symbols": ["Ti"], "concentration": [1.0]} ]`: any site with this species is occupied by a Ti atom.
        - :val:`[ {"name": "Ti", "chemical_symbols": ["Ti", "vacancy"], "concentration": [0.9, 0.1]} ]`: any site with this species is occupied by a Ti atom with 90 % probability, and has a vacancy with 10 % probability.
        - :val:`[ {"name": "BaCa", "chemical_symbols": ["vacancy", "Ba", "Ca"], "concentration": [0.05, 0.45, 0.5], "mass": 88.5} ]`: any site with this species is occupied by a Ba atom with 45 % probability, a Ca atom with 50 % probability, and by a vacancy with 5 % probability. The mass of this site is (on average) 88.5 a.m.u.
        - :val:`[ {"name": "C12", "chemical_symbols": ["C"], "concentration": [1.0], "mass": 12.0} ]`: any site with this species is occupied by a carbon isotope with mass 12.
        - :val:`[ {"name": "C13", "chemical_symbols": ["C"], "concentration": [1.0], "mass": 13.0} ]`: any site with this species is occupied by a carbon isotope with mass 13.

    """

    name: str = Field(
        ...,
        decsription="""REQUIRED; gives the name of the species; the **name** value MUST be unique in the :property:`species` list;""",
    )

    chemical_symbols: List[str] = Field(
        ...,
        description="""MUST be a list of strings of all chemical elements composing this species.

- It MUST be one of the following:

  - a valid chemical-element name, or
  - the special value :val:`"X"` to represent a non-chemical element, or
  - the special value :val:`"vacancy"` to represent that this site has a non-zero probability of having a vacancy (the respective probability is indicated in the :property:`concentration` list, see below).

-  If any one entry in the :property:`species` list has a :property:`chemical_symbols` list that is longer than 1 element, the correct flag MUST be set in the list :property:`structure_features` (see property `structure_features`_).""",
    )

    concentration: List[float] = Field(
        ...,
        description="""MUST be a list of floats, with same length as :property:`chemical_symbols`. The numbers represent the relative concentration of the corresponding chemical symbol in this species.
The numbers SHOULD sum to one. Cases in which the numbers do not sum to one typically fall only in the following two categories:

  - Numerical errors when representing float numbers in fixed precision, e.g. for two chemical symbols with concentrations :val:`1/3` and :val:`2/3`, the concentration might look something like :val:`[0.33333333333, 0.66666666666]`. If the client is aware that the sum is not one because of numerical precision, it can renormalize the values so that the sum is exactly one.
  - Experimental errors in the data present in the database. In this case, it is the responsibility of the client to decide how to process the data.

Note that concentrations are uncorrelated between different site (even of the same species).""",
    )

    mass: Optional[float] = Field(
        None,
        description="""If present MUST be a float expressed in a.m.u.""",
        unit="a.m.u.",
    )

    original_name: Optional[str] = Field(
        None,
        description="""Can be any valid Unicode string, and SHOULD contain (if specified) the name of the species that is used internally in the source database.

Note: With regards to "source database", we refer to the immediate source being queried via the OPTIMADE API implementation.
The main use of this field is for source databases that use species names, containing characters that are not allowed (see description of the list property `species_at_sites`_).""",
    )

    @validator("chemical_symbols", each_item=True)
    def validate_chemical_symbols(cls, v):
        if not (v in EXTENDED_CHEMICAL_SYMBOLS):
            raise ValueError(f"{v} MUST be in {EXTENDED_CHEMICAL_SYMBOLS}")
        return v

    @validator("concentration")
    def validate_concentration(cls, v, values):
        if len(v) != len(values.get("chemical_symbols", [])):
            raise ValueError(
                f"Length of concentration ({len(v)}) MUST equal length of chemical_symbols ({len(values.get('chemical_symbols', 'Not specified'))})"
            )
        return v


class Assembly(BaseModel):
    """A description of groups of sites that are statistically correlated.

    - **Examples** (for each entry of the assemblies list):

        - :val:`{"sites_in_groups": [[0], [1]], "group_probabilities: [0.3, 0.7]}`: the first site and the second site never occur at the same time in the unit cell.
            Statistically, 30 % of the times the first site is present, while 70 % of the times the second site is present.
        - :val:`{"sites_in_groups": [[1,2], [3]], "group_probabilities: [0.3, 0.7]}`: the second and third site are either present together or not present; they form the first group of atoms for this assembly.
            The second group is formed by the fourth site.
            Sites of the first group (the second and the third) are never present at the same time as the fourth site.
            30 % of times sites 1 and 2 are present (and site 3 is absent); 70 % of times site 3 is present (and sites 1 and 2 are absent).

    """

    sites_in_groups: List[List[int]] = Field(
        ...,
        description="""Index of the sites (0-based) that belong to each group for each assembly.

Example: :val:`[[1], [2]]`: two groups, one with the second site, one with the third.
Example: :val:`[[1,2], [3]]`: one group with the second and third site, one with the fourth.""",
    )

    group_probabilities: List[float] = Field(
        ...,
        description="""Statistical probability of each group. It MUST have the same length as :property:`sites_in_groups`.
It SHOULD sum to one.
See below for examples of how to specify the probability of the occurrence of a vacancy.
The possible reasons for the values not to sum to one are the same as already specified above for the :property:`concentration` of each :property:`species`, see property `species`_.""",
    )

    @validator("sites_in_groups")
    def validate_sites_in_groups(cls, v):
        sites = []
        for group in v:
            sites.extend(group)
        if len(set(sites)) != len(sites):
            raise ValueError(
                f"A site MUST NOT appear in more than one group. Given value: {v}"
            )
        return v

    @validator("group_probabilities")
    def check_self_consistency(cls, v, values):
        if len(v) != len(values.get("sites_in_groups", [])):
            raise ValueError(
                f"sites_in_groups and group_probabilities MUST be of same length, but are {len(values.get('sites_in_groups', 'Not specified'))} and {len(v)}, respectively"
            )
        return v


class StructureResourceAttributes(EntryResourceAttributes):
    """This class contains the Field for the attributes used to represent a structure, e.g. unit cell, atoms, positions."""

    elements: List[str] = Field(
        ...,
        description="""Names of the different elements present in the structure.
- **Type**: list of strings.
- **Requirements/Conventions**:

  - **Support**: SHOULD be supported, i.e., SHOULD NOT be :val:`null`. Is REQUIRED in this implementation, i.e., MUST NOT be :val:`null`.
  - **Query**: MUST be a queryable property with support for all mandatory filter operators.
  - The strings are the chemical symbols, written as uppercase letter plus optional lowercase letters.
  - The order MUST be alphabetical.
  - Note: This may not contain the "x" that is suggested in chemical_symbols for the :property:`species` property.

- **Examples**:

  - :val:`["Si"]`
  - :val:`["Al","O","Si"]`

- **Query examples**:
  - A filter that matches all records of structures that contain Si, Al **and** O, and possibly other elements: :filter:`elements HAS ALL "Si", "Al", "O"`.
  - To match structures with exactly these three elements, use :filter:`elements HAS ALL "Si", "Al", "O" AND LENGTH elements = 3`.""",
    )

    nelements: int = Field(
        ...,
        description="""Number of different elements in the structure as an integer.
- **Type**: integer
- **Requirements/Conventions**:

  - **Support**: SHOULD be supported, i.e., SHOULD NOT be :val:`null`. Is REQUIRED in this implementation, i.e., MUST NOT be :val:`null`.
  - **Query**: MUST be a queryable property with support for all mandatory filter operators.

- **Example**: :val:`3`
- **Querying**:

  -  Note: queries on this property can equivalently be formulated using :filter-fragment:`LENGTH elements`.
  -  A filter that matches structures that have exactly 4 elements: :filter:`nelements=4`.
  -  A filter that matches structures that have between 2 and 7 elements: :filter:`nelements>=2 AND nelements<=7`.""",
    )

    elements_ratios: List[float] = Field(
        ...,
        description="""Relative proportions of different elements in the structure.
- **Type**: list of floats
- **Requirements/Conventions**:

  - **Support**: SHOULD be supported, i.e., SHOULD NOT be :val:`null`. Is REQUIRED in this implementation, i.e., MUST NOT be :val:`null`.
  - **Query**: MUST be a queryable property with support for all mandatory filter operators.
  - Composed by the proportions of elements in the structure as a list of floating point numbers.
  - The sum of the numbers MUST be 1.0 (within floating point accuracy)

- **Examples**:

  - :val:`[1.0]`
  - :val:`[0.3333333333333333, 0.2222222222222222, 0.4444444444444444]`

- **Query examples**:

  - Note: useful filters can be formulated using the set operator syntax for correlated values. However, since the values are floating point values, the use of equality comparisons is generally not recommended.
  - A filter that matches structures where approximately 1/3 of the atoms in the structure are the element Al is: :filter:`elements:elements_ratios HAS ALL "Al":>0.3333, "Al":<0.3334`.""",
    )

    chemical_formula_descriptive: str = Field(
        ...,
        description="""The chemical formula for a structure as a string in a form chosen by the API implementation.
- **Type**: string
- **Requirements/Conventions**:

  - **Support**: SHOULD be supported, i.e., SHOULD NOT be :val:`null`. Is REQUIRED in this implementation, i.e., MUST NOT be :val:`null`.
  - **Query**: MUST be a queryable property with support for all mandatory filter operators.
  - The chemical formula is given as a string consisting of properly capitalized element symbols followed by integers or decimal numbers, balanced parentheses, square, and curly brackets ``(``, ``)``, ``[``, ``]``, ``{``, ``}``, commas, the ``+``, ``-``, ``:`` and ``=`` symbols.
    The parentheses are allowed to be followed by a number.
    Spaces are allowed anywhere except within chemical symbols.
    The order of elements and any groupings indicated by parentheses or brackets are chosen freely by the API implementation.
  - The string SHOULD be arithmetically consistent with the element ratios in the :property:`chemical_formula_reduced` property.
  - It is RECOMMENDED, but not mandatory, that symbols, parentheses and brackets, if used, are used with the meanings prescribed by `IUPAC's Nomenclature of Organic Chemistry <https://www.qmul.ac.uk/sbcs/iupac/bibliog/blue.html>`__.

- **Examples**:

  - :val:`"(H2O)2 Na"`
  - :val:`"NaCl"`
  - :val:`"CaCO3"`
  - :val:`"CCaO3"`
  - :val:`"(CH3)3N+ - [CH2]2-OH = Me3N+ - CH2 - CH2OH"`

- **Query examples**:

  - Note: the free-form nature of this property is likely to make queries on it across different databases inconsistent.
  - A filter that matches an exactly given formula: :filter:`chemical_formula_descriptive="(H2O)2 Na"`.
  - A filter that does a partial match: :filter:`chemical_formula_descriptive CONTAINS "H2O"`.""",
    )

    chemical_formula_reduced: str = Field(
        ...,
        description="""The reduced chemical formula for a structure as a string with element symbols and integer chemical proportion numbers.
  The proportion number MUST be omitted if it is 1.
- **Type**: string
- **Requirements/Conventions**:

  - **Support**: SHOULD be supported, i.e., SHOULD NOT be :val:`null`. Is REQUIRED in this implementation, i.e., MUST NOT be :val:`null`.
  - **Query**: MUST be a queryable property.
    However, support for filters using partial string matching with this property is OPTIONAL (i.e., BEGINS WITH, ENDS WITH, and CONTAINS).
    Intricate querying on formula components are instead recommended to be formulated using set-type filter operators on the multi valued :property:`elements` and :property:`elements_proportions` properties.
  - Element names MUST have proper capitalization (e.g., :val:`"Si"`, not :VAL:`"SI"` for "silicon").
  - Elements MUST be placed in alphabetical order, followed by their integer chemical proportion number.
  - For structures with no partial occupation, the chemical proportion numbers are the smallest integers for which the chemical proportion is exactly correct.
  - For structures with partial occupation, the chemical proportion numbers are integers that within reasonable approximation indicate the correct chemical proportions. The precise details of how to perform the rounding is chosen by the API implementation.
  - No spaces or separators are allowed.

- **Examples**:

  - :val:`"H2NaO"`
  - :val:`"ClNa"`
  - :val:`"CCaO3"`

- **Query examples**:

  - A filter that matches an exactly given formula is :filter:`chemical_formula_reduced="H2NaO"`.""",
    )

    chemical_formula_hill: Optional[str] = Field(
        None,
        description="""The chemical formula for a structure in `Hill form <https://dx.doi.org/10.1021/ja02046a005>`__ with element symbols followed by integer chemical proportion numbers.
  The proportion number MUST be omitted if it is 1.
- **Type**: string
- **Requirements/Conventions**:

  - **Support**: OPTIONAL, i.e., MAY be :val:`null`.
  - **Query**: Support for queries on these properties are OPTIONAL. If supported, only a subset of filter operators MAY be supported.
  - The overall scale factor of the chemical proportions is chosen such that the resulting values are integers that indicate the most chemically relevant unit of which the system is composed.
    For example, if the structure is a repeating unit cell with four hydrogens and four oxygens that represents two hydroperoxide molecules, :property:`chemical_formula_hill` is :val:`"H2O2"` (i.e., not :val:`"HO"`, nor :val:`"H4O4"`).
  - If the chemical insight needed to ascribe a Hill formula to the system is not present, the property MUST be handled as unset.
  - Element names MUST have proper capitalization (e.g., :val:`"Si"`, not :VAL:`"SI"` for "silicon").
  - Elements MUST be placed in `Hill order <https://dx.doi.org/10.1021/ja02046a005>`__, followed by their integer chemical proportion number.
    Hill order means: if carbon is present, it is placed first, and if also present, hydrogen is placed second.
    After that, all other elements are ordered alphabetically.
    If carbon is not present, all elements are ordered alphabetically.
  - If the system has sites with partial occupation and the total occupations of each element do not all sum up to integers, then the Hill formula SHOULD be handled as unset.
  - No spaces or separators are allowed.

- **Examples**:
  - :val:`"H2O2"`

- **Query examples**:

  - A filter that matches an exactly given formula is :filter:`chemical_formula_hill="H2O2"`.""",
    )

    chemical_formula_anonymous: str = Field(
        ...,
        description="""The anonymous formula is the :property:`chemical_formula_reduced`, but where the elements are instead first ordered by their chemical proportion number, and then, in order left to right, replaced by anonymous symbols A, B, C, ..., Z, Aa, Ba, ..., Za, Ab, Bb, ... and so on.
- **Type**: string
- **Requirements/Conventions**:

  - **Support**: SHOULD be supported, i.e., SHOULD NOT be :val:`null`. Is REQUIRED in this implementation, i.e., MUST NOT be :val:`null`.
  - **Query**: MUST be a queryable property. However, support for filters using partial string matching with this property is OPTIONAL (i.e., BEGINS WITH, ENDS WITH, and CONTAINS).

- **Examples**:

  - :val:`"A2B"`
  - :val:`"A42B42C16D12E10F9G5"`

- **Querying**:
  - A filter that matches an exactly given formula is :filter:`chemical_formula_anonymous="A2B"`.""",
    )

    dimension_types: Tuple[Periodicity, Periodicity, Periodicity] = Field(
        ...,
        description="""List of three integers.
  For each of the three directions indicated by the three lattice vectors (see property `lattice_vectors`_).
  This list indicates if the direction is periodic (value :val:`1`) or non-periodic (value :val:`0`).
  Note: the elements in this list each refer to the direction of the corresponding entry in property `lattice_vectors`_ and *not* the Cartesian x, y, z directions.
- **Type**: list of integers.
- **Requirements/Conventions**:

  - **Support**: SHOULD be supported, i.e., SHOULD NOT be :val:`null`. Is REQUIRED in this implementation, i.e., MUST NOT be :val:`null`.
  - **Query**: MUST be a queryable property. Support for equality comparison is REQUIRED, support for other comparison operators are OPTIONAL.
  - MUST be a list of length 3.
  - Each integer element MUST assume only the value 0 or 1.

- **Examples**:

  - For a molecule: :val:`[0, 0, 0]`
  - For a wire along the direction specified by the third lattice vector: :val:`[0, 0, 1]`
  - For a 2D surface/slab, periodic on the plane defined by the first and third lattice vectors: :val:`[1, 0, 1]`
  - For a bulk 3D system: :val:`[1, 1, 1]`""",
    )

    lattice_vectors: Optional[Tuple[Vector3D, Vector3D, Vector3D]] = Field(
        None,
        description="""The three lattice vectors in Cartesian coordinates, in ångström (Å).
- **Type**: list of list of floats.
- **Requirements/Conventions**:

  - **Support**: SHOULD be supported, i.e., SHOULD NOT be :val:`null`. Is REQUIRED in this implementation, i.e., MUST NOT be :val:`null`.
  - **Query**: Support for queries on this property is OPTIONAL.
    If supported, filters MAY support only a subset of comparison operators.
  - MUST be a list of three vectors *a*, *b*, and *c*, where each of the vectors MUST BE a list of the vector's coordinates along the x, y, and z Cartesian coordinates.
    (Therefore, the first index runs over the three lattice vectors and the second index runs over the x, y, z Cartesian coordinates).
  - For databases that do not define an absolute Cartesian system (e.g., only defining the length and angles between vectors), the first lattice vector SHOULD be set along *x* and the second on the *xy*-plane.
  - This property MUST be an array of dimensions 3 times 3 regardless of the elements of :property:`dimension_types`.
    The vectors SHOULD by convention be chosen so the determinant of the :property:`lattice_vectors` matrix is different from zero.
    The vectors in the non-periodic directions have no significance beyond fulfilling these requirements.
  - All three elements of the inner lists of floats MAY be :val:`null` for non-periodic dimensions, i.e., those dimensions for which :property:`dimension_types` is :val:`0`.

- **Examples**:

  - :val:`[[4.0,0.0,0.0],[0.0,4.0,0.0],[0.0,1.0,4.0]]` represents a cell, where the first vector is :val:`(4, 0, 0)`, i.e., a vector aligned along the :val:`x` axis of length 4 Å; the second vector is :val:`(0, 4, 0)`; and the third vector is :val:`(0, 1, 4)`.""",
        unit="Å",
    )

    cartesian_site_positions: List[Vector3D] = Field(
        ...,
        description="""Cartesian positions of each site. A site is an atom, a site potentially occupied by an atom, or a placeholder for a virtual mixture of atoms (e.g., in a virtual crystal approximation).
- **Type**: list of list of floats and/or unknown values
- **Requirements/Conventions**:

  - **Support**: SHOULD be supported, i.e., SHOULD NOT be :val:`null`. Is REQUIRED in this implementation, i.e., MUST NOT be :val:`null`.
  - **Query**: Support for queries on this property is OPTIONAL. If supported, filters MAY support only a subset of comparison operators.
  - It MUST be a list of length N times 3, where N is the number of sites in the structure.
  - An entry MAY have multiple sites at the same Cartesian position (for a relevant use of this, see e.g., the property `assemblies`_).
  - If a component of the position is unknown, the :val:`null` value should be provided instead (see section `Properties with unknown value`_).
    Otherwise, it should be a float value, expressed in angstrom (Å).
    If at least one of the coordinates is unknown, the correct flag in the list property `structure_features`_ MUST be set.
  - **Notes**: (for implementers) While this is unrelated to this OPTIMADE specification: If you decide to store internally the :property: `cartesian_site_positions` as a float array, you might want to represent :val:`null` values with :field-val:`NaN` values.
    The latter being valid float numbers in the IEEE 754 standard in `IEEE 754-1985 <https://doi.org/10.1109/IEEESTD.1985.82928>`__ and in the updated version `IEEE 754-2008 <https://doi.org/10.1109/IEEESTD.2008.4610935>`__.

- **Examples**:

  - :val:`[[0,0,0],[0,0,2]]` indicates a structure with two sites, one sitting at the origin and one along the (positive) *z*-axis, 2 Å away from the origin.""",
        unit="Å",
    )

    nsites: int = Field(
        ...,
        description="""An integer specifying the length of the :property:`cartesian_site_positions` property.
- **Type**: integer
- **Requirements/Conventions**:

  - **Support**: SHOULD be supported, i.e., SHOULD NOT be :val:`null`. Is REQUIRED in this implementation, i.e., MUST NOT be :val:`null`.
  - **Query**: MUST be a queryable property with support for all mandatory filter operators.

- **Examples**:

  - :val:`42`

- **Query examples**:

  - Match only structures with exactly 4 sites: :filter:`nsites=4`
  - Match structures that have between 2 and 7 sites: :filter:`nsites>=2 AND nsites<=7`""",
    )

    species_at_sites: List[str] = Field(
        ...,
        description="""Name of the species at each site (where values for sites are specified with the same order of the property `cartesian_site_positions`_).
  The properties of the species are found in the property `species`_.
- **Type**: list of strings.
- **Requirements/Conventions**:

  - **Support**: SHOULD be supported, i.e., SHOULD NOT be :val:`null`. Is REQUIRED in this implementation, i.e., MUST NOT be :val:`null`.
  - **Query**: Support for queries on this property is OPTIONAL. If supported, filters MAY support only a subset of comparison operators.
  - MUST have length equal to the number of sites in the structure (first dimension of the list property `cartesian_site_positions`_).
  - Each species MUST have a unique name.
  - Each species name mentioned in the :property:`species_at_sites` list MUST be described in the list property `species`_ (i.e. for each value in the :property:`species_at_sites` list there MUST exist exactly one dictionary in the :property:`species` list with the :property:`name` attribute equal to the corresponding :property:`species_at_sites` value).
  - Each site MUST be associated only to a single species.
    **Note**: However, species can represent mixtures of atoms, and multiple species MAY be defined for the same chemical element.
    This latter case is useful when different atoms of the same type need to be grouped or distinguished, for instance in simulation codes to assign different initial spin states.

- **Examples**:

  - :val:`["Ti","O2"]` indicates that the first site is hosting a species labeled :val:`"Ti"` and the second a species labeled :val:`"O2"`.""",
    )

    species: List[Species] = Field(
        ...,
        description="""A list describing the species of the sites of this structure. Species can be pure chemical elements, or virtual-crystal atoms representing a statistical occupation of a given site by multiple chemical elements.
- **Type**: list of dictionary with keys:

  - :property:`name`: string (REQUIRED)
  - :property:`chemical_symbols`: list of strings (REQUIRED)
  - :property:`concentration`: list of float (REQUIRED)
  - :property:`mass`: float (OPTIONAL)
  - :property:`original_name`: string (OPTIONAL).

- **Requirements/Conventions**:

  - **Support**: SHOULD be supported, i.e., SHOULD NOT be :val:`null`. Is REQUIRED in this implementation, i.e., MUST NOT be :val:`null`.
  - **Query**: Support for queries on this property is OPTIONAL.
    If supported, filters MAY support only a subset of comparison operators.
  - Each list member MUST be a dictionary with the following keys:

    - **name**: REQUIRED; gives the name of the species; the **name** value MUST be unique in the :property:`species` list;

    - **chemical_symbols**: REQUIRED; MUST be a list of strings of all chemical elements composing this species.

      - It MUST be one of the following:

        - a valid chemical-element name, or
        - the special value :val:`"X"` to represent a non-chemical element, or
        - the special value :val:`"vacancy"` to represent that this site has a non-zero probability of having a vacancy (the respective probability is indicated in the :property:`concentration` list, see below).

      -  If any one entry in the :property:`species` list has a :property:`chemical_symbols` list that is longer than 1 element, the correct flag MUST be set in the list :property:`structure_features` (see property `structure_features`_).

    - **concentration**: REQUIRED; MUST be a list of floats, with same length as :property:`chemical_symbols`. The numbers represent the relative concentration of the corresponding chemical symbol in this species.
      The numbers SHOULD sum to one. Cases in which the numbers do not sum to one typically fall only in the following two categories:

      - Numerical errors when representing float numbers in fixed precision, e.g. for two chemical symbols with concentrations :val:`1/3` and :val:`2/3`, the concentration might look something like :val:`[0.33333333333, 0.66666666666]`. If the client is aware that the sum is not one because of numerical precision, it can renormalize the values so that the sum is exactly one.
      - Experimental errors in the data present in the database. In this case, it is the responsibility of the client to decide how to process the data.

      Note that concentrations are uncorrelated between different site (even of the same species).

    - **mass**: OPTIONAL. If present MUST be a float expressed in a.m.u.
    - **original_name**: OPTIONAL. Can be any valid Unicode string, and SHOULD contain (if specified) the name of the species that is used internally in the source database.

        Note: With regards to "source database", we refer to the immediate source being queried via the OPTIMADE API implementation.
            The main use of this field is for source databases that use species names, containing characters that are not allowed (see description of the list property `species_at_sites`_).

  - For systems that have only species formed by a single chemical symbol, and that have at most one species per chemical symbol, SHOULD use the chemical symbol as species name (e.g., :val:`"Ti"` for titanium, :val:`"O"` for oxygen, etc.)
    However, note that this is OPTIONAL, and client implementations MUST NOT assume that the key corresponds to a chemical symbol, nor assume that if the species name is a valid chemical symbol, that it represents a species with that chemical symbol.
    This means that a species :val:`{"name": "C", "chemical_symbols": ["Ti"], "concentration": [1.0]}` is valid and represents a titanium species (and *not* a carbon species).
  - It is NOT RECOMMENDED that a structure includes species that do not have at least one corresponding site.

- **Examples**:

  - :val:`[ {"name": "Ti", "chemical_symbols": ["Ti"], "concentration": [1.0]} ]`: any site with this species is occupied by a Ti atom.
  - :val:`[ {"name": "Ti", "chemical_symbols": ["Ti", "vacancy"], "concentration": [0.9, 0.1]} ]`: any site with this species is occupied by a Ti atom with 90 % probability, and has a vacancy with 10 % probability.
  - :val:`[ {"name": "BaCa", "chemical_symbols": ["vacancy", "Ba", "Ca"], "concentration": [0.05, 0.45, 0.5], "mass": 88.5} ]`: any site with this species is occupied by a Ba atom with 45 % probability, a Ca atom with 50 % probability, and by a vacancy with 5 % probability. The mass of this site is (on average) 88.5 a.m.u.
  - :val:`[ {"name": "C12", "chemical_symbols": ["C"], "concentration": [1.0], "mass": 12.0} ]`: any site with this species is occupied by a carbon isotope with mass 12.
  - :val:`[ {"name": "C13", "chemical_symbols": ["C"], "concentration": [1.0], "mass": 13.0} ]`: any site with this species is occupied by a carbon isotope with mass 13.""",
    )

    assemblies: Optional[List[Assembly]] = Field(
        None,
        description="""A description of groups of sites that are statistically correlated.
- **Type**: list of dictionary with keys:

  - :property:`sites_in_groups`: list of list of integers (REQUIRED)
  - :property:`group_probabilities`: list of floats (REQUIRED)

- **Requirements/Conventions**:

  - **Support**: OPTIONAL support in implementations, i.e., MAY be :val:`null`.
  - **Query**: Support for queries on this property is OPTIONAL.
    If supported, filters MAY support only a subset of comparison operators.
  - If present, the correct flag MUST be set in the list :property:`structure_features` (see property `structure_features`_).
  - Client implementations MUST check its presence (as its presence changes the interpretation of the structure).
  - If present, it MUST be a list of dictionaries, each of which represents an assembly and MUST have the following two keys:

    - **sites_in_groups**: Index of the sites (0-based) that belong to each group for each assembly.

      Example: :val:`[[1], [2]]`: two groups, one with the second site, one with the third.

      Example: :val:`[[1,2], [3]]`: one group with the second and third site, one with the fourth.

   - **group_probabilities**: Statistical probability of each group. It MUST have the same length as :property:`sites_in_groups`.
     It SHOULD sum to one.
     See below for examples of how to specify the probability of the occurrence of a vacancy.
     The possible reasons for the values not to sum to one are the same as already specified above for the :property:`concentration` of each :property:`species`, see property `species`_.

  - If a site is not present in any group, it means that it is present with 100 % probability (as if no assembly was specified).
  - A site MUST NOT appear in more than one group.

- **Examples** (for each entry of the assemblies list):

  - :val:`{"sites_in_groups": [[0], [1]], "group_probabilities: [0.3, 0.7]}`: the first site and the second site never occur at the same time in the unit cell.
    Statistically, 30 % of the times the first site is present, while 70 % of the times the second site is present.
  - :val:`{"sites_in_groups": [[1,2], [3]], "group_probabilities: [0.3, 0.7]}`: the second and third site are either present together or not present; they form the first group of atoms for this assembly.
    The second group is formed by the fourth site.
    Sites of the first group (the second and the third) are never present at the same time as the fourth site.
    30 % of times sites 1 and 2 are present (and site 3 is absent); 70 % of times site 3 is present (and sites 1 and 2 are absent).

- **Notes**:

  - Assemblies are essential to represent, for instance, the situation where an atom can statistically occupy two different positions (sites).
  - By defining groups, it is possible to represent, e.g., the case where a functional molecule (and not just one atom) is either present or absent (or the case where it it is present in two conformations)
  - Considerations on virtual alloys and on vacancies: In the special case of a virtual alloy, these specifications allow two different, equivalent ways of specifying them.
    For instance, for a site at the origin with 30 % probability of being occupied by Si, 50 % probability of being occupied by Ge, and 20 % of being a vacancy, the following two representations are possible:

    - Using a single species:

      .. code:: jsonc

           {
             "cartesian_site_positions": [[0,0,0]],
             "species_at_sites": ["SiGe-vac"],
             "species": [
                 {
                   "name": "SiGe-vac",
                   "chemical_symbols": ["Si", "Ge", "vacancy"],
                   "concentration": [0.3, 0.5, 0.2]
                 }
             ]
             // ...
           }


    - Using multiple species and the assemblies:

      .. code:: jsonc

           {
             "cartesian_site_positions": [ [0,0,0], [0,0,0], [0,0,0] ],
             "species_at_sites": ["Si", "Ge", "vac"],
             "species": {
               "Si": { "chemical_symbols": ["Si"], "concentration": [1.0] },
               "Ge": { "chemical_symbols": ["Ge"], "concentration": [1.0] },
               "vac": { "chemical_symbols": ["vacancy"], "concentration": [1.0] }
             },
             "assemblies": [
               {
                 "sites_in_groups": [ [0], [1], [2] ],
                 "group_probabilities": [0.3, 0.5, 0.2]
               }
             ]
             // ...
           }

  - It is up to the database provider to decide which representation to use, typically depending on the internal format in which the structure is stored.
    However, given a structure identified by a unique ID, the API implementation MUST always provide the same representation for it.
  - The probabilities of occurrence of different assemblies are uncorrelated.
    So, for instance in the following case with two assemblies:

    .. code:: jsonc

         {
           "assemblies": [
             {
               "sites_in_groups": [ [0], [1] ],
               "group_probabilities": [0.2, 0.8],
             },
             {
               "sites_in_groups": [ [2], [3] ],
               "group_probabilities": [0.3, 0.7]
             }
           ]
         }

    Site 0 is present with a probability of 20 % and site 1 with a probability of 80 %. These two sites are correlated (either site 0 or 1 is present). Similarly, site 2 is present with a probability of 30 % and site 3 with a probability of 70 %.
    These two sites are correlated (either site 2 or 3 is present).
    However, the presence or absence of sites 0 and 1 is not correlated with the presence or absence of sites 2 and 3 (in the specific example, the pair of sites (0, 2) can occur with 0.2*0.3 = 6 % probability; the pair (0, 3) with 0.2*0.7 = 14 % probability; the pair (1, 2) with 0.8*0.3 = 24 % probability; and the pair (1, 3) with 0.8*0.7 = 56 % probability).""",
    )

    structure_features: List[str] = Field(
        ...,
        description="""A list of strings that flag which special features are used by the structure.
- **Type**: list of strings
- **Requirements/Conventions**:

  - **Support**: REQUIRED, MUST NOT be :val:`null`.
  - **Query**: MUST be a queryable property. Filters on the list MUST support all mandatory HAS-type queries. Filter operators for comparisons on the string components MUST support equality, support for other comparison operators are OPTIONAL.
  - MUST be an empty list if no special features are used.
  - MUST be sorted alphabetically.
  - If a special feature listed below is used, the list MUST contain the corresponding string.
  - If a special feature listed below is not used, the list MUST NOT contain the corresponding string.
  - **List of strings used to indicate special structure features**:

    - :val:`disorder`: This flag MUST be present if any one entry in the :property:`species` list has a :property:`chemical_symbols` list that is longer than 1 element.
    - :val:`unknown_positions`: This flag MUST be present if at least one component of the :property:`cartesian_site_positions` list of lists has value :val:`null`.
    - :val:`assemblies`: This flag MUST be present if the property `assemblies`_ is present.

-  **Examples**: A structure having unknown positions and using assemblies: :val:`["assemblies", "unknown_positions"]`""",
    )

    @validator("elements", each_item=True)
    def element_must_be_chemical_symbol(cls, v):
        if v not in CHEMICAL_SYMBOLS:
            raise ValueError(f"Only chemical symbols are allowed, you passed: {v}")
        return v

    @validator("elements")
    def elements_must_be_alphabetical(cls, v):
        if sorted(v) != v:
            raise ValueError(f"elements must be sorted alphabetically, but is: {v}")
        return v

    @validator("elements_ratios")
    def ratios_must_sum_to_one(cls, v):
        if abs(sum(v) - 1) > EPS:
            raise ValueError(
                f"elements_ratios MUST sum to 1 within floating point accuracy. It sums to: {sum(v)}"
            )
        return v

    @validator("chemical_formula_reduced", "chemical_formula_hill")
    def no_spaces_in_reduced(cls, v):
        if v and " " in v:
            raise ValueError(f"Spaces are not allowed, you passed: {v}")
        return v

    @validator("lattice_vectors", always=True)
    def required_if_dimension_types_has_one(cls, v, values):
        if 1 in values.get("dimension_types", []) and v is None:
            raise ValueError(
                f"lattice_vectors is REQUIRED, since dimension_types is not [0, 0, 0] but is {values.get('dimension_types', 'Not specified')}"
            )

        for dim_type, vector in zip(values.get("dimension_types", []), v):
            if None in vector and dim_type == 1:
                raise ValueError(
                    "Null entries in lattice vectors are only permitted when the corresponding dimension type is 0. "
                    f"Here: dimension_types = {values.get('dimension_types', 'Not specified')}, lattice_vectors = {v}"
                )

        return v

    @validator("nsites")
    def validate_nsites(cls, v, values):
        if v != len(values.get("cartesian_site_positions", [])):
            raise ValueError(
                f"nsites (value: {v}) MUST equal length of cartesian_site_positions (value: {len(values.get('cartesian_site_positions', []))})"
            )
        return v

    @validator("species_at_sites")
    def validate_species_at_sites(cls, v, values):
        if "nsites" not in values:
            raise ValueError(
                "Attribute nsites missing so unable to verify species_at_sites."
            )
        if len(v) != values.get("nsites", 0):
            raise ValueError(
                f"Number of species_at_sites (value: {len(v)}) MUST equal number of sites (value: {values.get('nsites', 'Not specified')})"
            )
        return v

    @validator("species", each_item=True)
    def validate_species(cls, v, values):
        if v.name not in values.get("species_at_sites", []):
            raise ValueError(
                f"{v.name} not found in species_at_sites: {values.get('species_at_sites', 'Not specified')}"
            )
        return v

    @validator("structure_features", always=True)
    def validate_structure_features(cls, v, values):
        if sorted(v) != v:
            raise ValueError(
                f"structure_features MUST be sorted alphabetically, given value: {v}"
            )
        # disorder
        for species in values.get("species", []):
            if len(species.chemical_symbols) > 1:
                if "disorder" not in v:
                    raise ValueError(
                        "disorder MUST be present when any one entry in species has a chemical_symbols list greater than one element"
                    )
                break
        else:
            if "disorder" in v:
                raise ValueError(
                    "disorder MUST NOT be present, since all species' chemical_symbols lists are equal to or less than one element"
                )
        # unknown_positions
        for site in values.get("cartesian_site_positions", []):
            if None in site or float("nan") in site:
                if "unknown_positions" not in v:
                    raise ValueError(
                        "unknown_positions MUST be present when a single component of cartesian_site_positions has value null"
                    )
                break
        else:
            if "unknown_positions" in v:
                raise ValueError(
                    "unknown_positions MUST NOT be present, since there are no null values in cartesian_site_positions"
                )
        # assemblies
        if values.get("assemblies", None) is not None:
            if "assemblies" not in v:
                raise ValueError(
                    "assemblies MUST be present, since the property of the same name is present"
                )
        else:
            if "assemblies" in v:
                raise ValueError(
                    "assemblies MUST NOT be present, since the property of the same name is not present"
                )
        return v


class StructureResource(EntryResource):
    """Representing a structure."""

    type: str = Field(
        "structures",
        const=True,
        description="""The name of the type of an entry. Any entry MUST be able to be fetched using the `base URL <Base URL_>`_ type and ID at the url :endpoint:`<base URL>/<type>/<id>`.
- **Type**: string.
- **Requirements/Conventions**:

  - **Support**: REQUIRED, MUST NOT be :val:`null`.
  - **Query**: MUST be a queryable property with support for all mandatory filter features.
  - **Response**: REQUIRED in the response.
  - MUST be an existing entry type.
  - The entry of type `<type>` and ID `<id>` MUST be returned in response to a request for :endpoint:`/<type>/<id>` under the versioned base URL.

- **Example**: :val:`"structures"`""",
    )

    attributes: StructureResourceAttributes



from fastapi import Query
from pydantic import EmailStr  # pylint: disable=no-name-in-module

from .config import CONFIG


class EntryListingQueryParams:
    """Common query params for all Entry listing endpoints."""

    def __init__(
        self,
        *,
        filter: str = Query(  # pylint: disable=redefined-builtin
            "",
            description="A filter string, in the format described in section [API Filtering Format Specification]"
            "(https://github.com/Materials-Consortia/OPTIMADE/blob/develop/optimade.rst#api-filtering-format-specification) "
            "of the [OPTIMADE spec](https://github.com/Materials-Consortia/OPTIMADE/blob/develop/optimade.rst).",
        ),
        response_format: str = Query(
            "json",
            description="The output format requested (see section [Response Format](https://github.com/Materials-Consortia/OPTIMADE/blob/develop/"
            "optimade.rst#response-format) in the spec). Defaults to the format string 'json', which specifies the standard output format "
            "described in this specification.\n**Example**: http://example.com/v1/structures?response_format=xml",
        ),
        email_address: EmailStr = Query(
            "",
            description="An email address of the user making the request. The email SHOULD be that of a person and not an automatic system.\n"
            "**Example**: http://example.com/v1/structures?email_address=user@example.com",
        ),
        response_fields: str = Query(
            "",
            description="A comma-delimited set of fields to be provided in the output. If provided, these fields MUST be returned along with "
            "the REQUIRED fields. Other OPTIONAL fields MUST NOT be returned when this parameter is present.\n"
            "**Example**: http://example.com/v1/structures?response_fields=last_modified,nsites",
            regex=r"([a-z_][a-z_0-9]*(,[a-z_][a-z_0-9]*)*)?",
        ),
        sort: str = Query(
            "",
            description="If supporting sortable queries, an implementation MUST use the `sort` query parameter with format as specified by "
            "[JSON API 1.0](https://jsonapi.org/format/1.0/#fetching-sorting).\n\n"
            "An implementation MAY support multiple sort fields for a single query. If it does, it again MUST conform to the JSON API 1.0 specification.\n\n"
            "If an implementation supports sorting for an [entry listing endpoint](https://github.com/Materials-Consortia/OPTIMADE/blob/develop/optimade.rst"
            "#entry-listing-endpoints), then the `/info/<entries>` endpoint MUST include, for each field name `<fieldname>` in its "
            "`data.properties.<fieldname>` response value that can be used for sorting, the key `sortable` with value `true`. "
            "If a field name under an entry listing endpoint supporting sorting cannot be used for sorting, the server MUST either leave out the `sortable` "
            "key or set it equal to `false` for the specific field name. The set of field names, with `sortable` equal to `true` are allowed to be used in "
            'the "sort fields" list according to its definition in the JSON API 1.0 specification. The field `sortable` is in addition to each property '
            "description (and the OPTIONAL field `unit`). An example is shown in section "
            "[Entry Listing Info Endpoints](https://github.com/Materials-Consortia/OPTIMADE/blob/develop/optimade.rst#entry-listing-info-endpoints).",
            regex=r"([a-z_][a-z_0-9]*(,[a-z_][a-z_0-9]*)*)?",
        ),
        page_limit: int = Query(
            CONFIG.page_limit,
            description="Sets a numerical limit on the number of entries returned. "
            "See [JSON API 1.0](https://jsonapi.org/format/1.0/#fetching-pagination). The API implementation MUST return no more than the number specified. "
            "It MAY return fewer. The database MAY have a maximum limit and not accept larger numbers (in which case an error code -- 403 Forbidden -- "
            "MUST be returned). The default limit value is up to the API implementation to decide.\n\n"
            "A server MUST implement pagination in the case of no user-specified `sort` parameter (via the `links` response field, see section "
            "[JSON Response Schema: Common Fields](https://github.com/Materials-Consortia/OPTIMADE/blob/develop/optimade.rst#json-response-schema-"
            "common-fields)). A server MAY implement pagination in concert with `sort`.",
            ge=0,
        ),
        page_offset: int = Query(
            0,
            description="RECOMMENDED for use with _offset-based_ pagination: using `page_offset` and `page_limit` is RECOMMENDED.\n"
            "**Example**: Skip 50 structures and fetch up to 100: `/structures?page_offset=50&page_limit=100`.",
            ge=0,
        ),
        page_number: int = Query(
            0,
            description="RECOMMENDED for use with _page-based_ pagination: using `page_number` and `page_limit` is RECOMMENDED. "
            "It is RECOMMENDED that the first page has number 1, i.e., that `page_number` is 1-based.\n"
            "**Example**: Fetch page 2 of up to 50 structures per page: `/structures?page_number=2&page_limit=50`.",
            ge=0,
        ),
        page_cursor: int = Query(
            0,
            description="RECOMMENDED for use with _cursor-based_ pagination: using `page_cursor` and `page_limit` is RECOMMENDED.",
            ge=0,
        ),
        page_above: int = Query(
            0,
            description="RECOMMENDED for use with _value-based_ pagination: using `page_above`/`page_below` and `page_limit` is RECOMMENDED.\n"
            "**Example**: Fetch up to 100 structures above sort-field value 4000 (in this example, server chooses to fetch results sorted by increasing id, "
            "so `page_above` value refers to an `id` value): `/structures?page_above=4000&page_limit=100`.",
            ge=0,
        ),
        page_below: int = Query(
            0,
            description="RECOMMENDED for use with _value-based_ pagination: using `page_above`/`page_below` and `page_limit` is RECOMMENDED.",
            ge=0,
        ),
        include: str = Query(
            "references",
            description="The JSON API concept of returning [compound documents](https://jsonapi.org/format/1.0/#document-compound-documents) "
            "by utilizing the `include` query parameter as specified by [JSON API 1.0](https://jsonapi.org/format/1.0/#fetching-includes).\n\n"
            "All related resource objects MUST be returned as part of an array value for the top-level `included` field, see section "
            "[JSON Response Schema: Common Fields](https://github.com/Materials-Consortia/OPTIMADE/blob/develop/optimade.rst#json-response-"
            "schema-common-fields).\n\n"
            'The value of `include` MUST be a comma-separated list of "relationship paths", as defined in the '
            "[JSON API](https://jsonapi.org/format/1.0/#fetching-includes). If relationship paths are not supported, or a server is unable "
            "to identify a relationship path a `400 Bad Request` response MUST be made.\n\n"
            "The **default value** for `include` is `references`. This means `references` entries MUST always be included under the top-level "
            "field `included` as default, since a server assumes if `include` is not specified by a client in the request, it is still specified "
            "as `include=references`. Note, if a client explicitly specifies `include` and leaves out `references`, `references` resource objects "
            "MUST NOT be included under the top-level field `included`, as per the definition of `included`, see section "
            "[JSON Response Schema: Common Fields](https://github.com/Materials-Consortia/OPTIMADE/blob/develop/optimade.rst#json-response-"
            "schema-common-fields).\n\n"
            "- **Note**: A query with the parameter `include` set to the empty string means no related resource objects are to be returned under the "
            "top-level field `included`.",
        ),
    ):
        self.filter = filter
        self.response_format = response_format
        self.email_address = email_address
        self.response_fields = response_fields
        self.sort = sort
        self.page_limit = page_limit
        self.page_offset = page_offset
        self.page_number = page_number
        self.page_cursor = page_cursor
        self.page_above = page_above
        self.page_below = page_below
        self.include = include


class SingleEntryQueryParams:
    """Common query params for single entry endpoints."""

    def __init__(
        self,
        *,
        response_format: str = Query(
            "json",
            description="The output format requested (see section [Response Format](https://github.com/Materials-Consortia/OPTIMADE/blob/develop/"
            "optimade.rst#response-format) in the spec). Defaults to the format string 'json', which specifies the standard output format "
            "described in this specification.\n**Example**: http://example.com/v1/structures?response_format=xml",
        ),
        email_address: EmailStr = Query(
            "",
            description="An email address of the user making the request. The email SHOULD be that of a person and not an automatic system.\n"
            "**Example**: http://example.com/v1/structures?email_address=user@example.com",
        ),
        response_fields: str = Query(
            "",
            description="A comma-delimited set of fields to be provided in the output. If provided, these fields MUST be returned along with "
            "the REQUIRED fields. Other OPTIONAL fields MUST NOT be returned when this parameter is present.\n"
            "**Example**: http://example.com/v1/structures?response_fields=last_modified,nsites",
            regex=r"([a-z_][a-z_0-9]*(,[a-z_][a-z_0-9]*)*)?",
        ),
        include: str = Query(
            "references",
            description="The JSON API concept of returning [compound documents](https://jsonapi.org/format/1.0/#document-compound-documents) "
            "by utilizing the `include` query parameter as specified by [JSON API 1.0](https://jsonapi.org/format/1.0/#fetching-includes).\n\n"
            "All related resource objects MUST be returned as part of an array value for the top-level `included` field, see section "
            "[JSON Response Schema: Common Fields](https://github.com/Materials-Consortia/OPTIMADE/blob/develop/optimade.rst#json-response-"
            "schema-common-fields).\n\n"
            'The value of `include` MUST be a comma-separated list of "relationship paths", as defined in the '
            "[JSON API](https://jsonapi.org/format/1.0/#fetching-includes). If relationship paths are not supported, or a server is unable "
            "to identify a relationship path a `400 Bad Request` response MUST be made.\n\n"
            "The **default value** for `include` is `references`. This means `references` entries MUST always be included under the top-level "
            "field `included` as default, since a server assumes if `include` is not specified by a client in the request, it is still specified "
            "as `include=references`. Note, if a client explicitly specifies `include` and leaves out `references`, `references` resource objects "
            "MUST NOT be included under the top-level field `included`, as per the definition of `included`, see section "
            "[JSON Response Schema: Common Fields](https://github.com/Materials-Consortia/OPTIMADE/blob/develop/optimade.rst#json-response-"
            "schema-common-fields).\n\n"
            "- **Note**: A query with the parameter `include` set to the empty string means no related resource objects are to be returned under the "
            "top-level field `included`.",
        ),
    ):
        self.response_format = response_format
        self.email_address = email_address
        self.response_fields = response_fields
        self.include = include



from __future__ import print_function
from past.builtins import cmp
from builtins import object
from sqlalchemy import Table, Column, Integer, String, Text, Unicode, UnicodeText, DateTime, ForeignKey, Boolean, SmallInteger, Enum, Float, and_
from sqlalchemy.orm import relationship, aliased
from sqlalchemy.orm.session import object_session
from sqlalchemy.sql.expression import insert, delete, select

import datetime, logging, json
from collections import namedtuple

from nomadcore.local_info_kinds import InfoKindEl, InfoKindEnv
from nomadcore.basic_meta_info import allMetaInfo, metaMetaInfo
from nomadcore.utils import goInteractive

from nomadcore.model_base import Base, useJson, compareSorted
if useJson:
    from nomadcore.model_base import JSON, JSONB

gidSize = 39
# hash  sizes(hexencoded) speed(with 64bit HW)
# md4    32    1 25%-50% more than md5
# md5    32    2 ~2 times sha256
# sha224 56    4
# sha256 64    4
# sha384 96    3 ~1.2-1.5 times sha224
# sha512 128   3
# sha1   40    not worth using, slower than sha256 in optimized implementations

# sha224 is the basis of the gid, then a character is added at the start depending on the
# a lower case character is prepended to indicate the method used to calculate the sha:
# u: uuid
# f: content of a file
# d: recursive sha of a directory
# p: a property sha
# i: dataTypeSha + standard formatted json data + propertyShas
# h: reproducible sha of hdf5 data
# j: sha of standard formatted json data
# t: timestamp

# ?: assimilation sha (prefix is the same as directory)
# s: a symbolic pseudo sha
# a: access group sha
# a: access control group sha
# T: timestampBlock
#
#
# A second character uppercase character might be prepended to build the egid:
# I: Info
# K: InfoKind
# A: AccessGroup or AccessControlGroup
# T: Trusted info (self generated, or vetted info)
# this lets the gid be trivially the sha of the content of the file without limitations
# or extra strings, while avoiding clashes arising from crafting files containing
# what is hashed in the other cases.
# The second character ensures that it is impossible to tamper with the separate
# access control no matter what one manags to store in the Info table.

info_kind_parents = Table('info_kind_parents', Base.metadata,
    Column('info_kind_gid', Integer, ForeignKey('info_kinds.gid'), nullable=False),
    Column('parent_gid', Integer, ForeignKey('info_kinds.gid'), nullable=False)
)
info_kind_parent = namedtuple('info_kind_parent', ['info_kind_gid', 'parent_gid'])

info_kind_anchestors = Table('info_kind_anchestors', Base.metadata,
    Column('info_kind_gid', Integer, ForeignKey('info_kinds.gid'), nullable=False),
    Column('anchestor_gid', Integer, ForeignKey('info_kinds.gid'), nullable=False)
)
info_kind_anchestor = namedtuple('info_kind_anchestor', ['info_kind_gid', 'anchestor_gid'])

class UnsetValue(object):
    pass

class InfoKind(Base):
    """Represents meta information, i.e. information on the information

    This is like a more powerful and structured set of tags.
    It is used both to define the type of information, and the pieces of information
    contained in it.

    It is always public, and versioning on them is also public (public_override)
    and thus has to be decided (vetted) by admins.

    There are 4 root types:
    - MetaType: implicit kind of the root InfoKind that describe InfoKind.kind.
        identified by a null InfoKind.kind
    - DocumentType: kind of InfoKinds describing the type of documents (Infos)
    - DocumentContentType: kind of InfoKinds describing the type of information contained
        in documents (Infos)
    - ConnectionType: kind of InfoKinds describing Connections
    - UnknownType: kind for the unknown InfoKinds kinds (only the name of them is relevant,
        to recover the correct kindStr)
    """
    __tablename__="info_kinds"
    gid = Column(String(gidSize), primary_key=True)
    name = Column(Unicode(128))
    description = Column(UnicodeText)
    kind_gid = Column(String(gidSize), ForeignKey('info_kinds.gid'), nullable=True)
    kind = relationship('InfoKind', foreign_keys=[kind_gid], remote_side = [gid])
    #meta_childs = relationship('InfoKind', foreign_keys=[kind_gid], remote_side=[gid])
    units = Column(Unicode(128))
    super_kinds = relationship("InfoKind",
                                                secondary=info_kind_parents,
                                                primaryjoin=gid==info_kind_parents.c.info_kind_gid,
                                                secondaryjoin=gid==info_kind_parents.c.parent_gid,
                                                backref="sub_kinds")
    super_kind_gid=Column(String(gidSize),ForeignKey('info_kinds.gid'), nullable=True)
    super_kind=relationship("InfoKind",foreign_keys=[super_kind_gid])
    cumulative_super_kinds = relationship("InfoKind",
                                                secondary=info_kind_anchestors,
                                                primaryjoin=gid==info_kind_anchestors.c.info_kind_gid,
                                                secondaryjoin=gid==info_kind_anchestors.c.anchestor_gid,
                                                backref="cumulative_sub_kinds")
    dtypeStr = Column(Unicode(128))
    repeats = Column(Boolean, nullable = True)
    if useJson:
        shape = Column(JSONB)
        extraArgs = Column(JSONB)
    else:
        shape_txt = Column(UnicodeText)
        extra_args_txt = Column(UnicodeText)
        def shape():
            doc = "The shape of array that stores this property (relevant for 'simple' properties)."
            def fget(self):
                if self.shape_txt:
                    return json.loads(self.shape_txt)
                return None
            def fset(self, value):
                self.shape_txt = json.dumps(value)
            return locals()
        shape = property(**shape())
        def extra_args():
            doc = "The extra_args of this InfoKinds."
            def fget(self):
                if self.extra_args_txt:
                    return json.loads(self.extra_args_txt)
                return None
            def fset(self, value):
                if value:
                    self.extra_args_txt = json.dumps(value).decode('utf-8')
            return locals()
        extra_args = property(**extra_args())
    public = Column(Boolean, default = True)
    public_override_gid = Column(String(gidSize), ForeignKey('info_kinds.gid'), nullable=True)
    public_override = relationship("InfoKind", foreign_keys=[public_override_gid])
    superseded_kinds = relationship("InfoKind", foreign_keys=[public_override_gid], remote_side=[gid])

    def super_gids():
        doc = "The super_gids property."
        def fget(self):
            res = [x.parent_gid for x in object_session(self).query(info_kind_parents.c.parent_gid).filter(info_kind_parents.c.info_kind_gid == self.gid).order_by(info_kind_parents.c.parent_gid).all()]
            if self.super_kind_gid:
                if self.super_kind_gid in res:
                    res.remove(self.super_kind_gid)
                res.insert(0, self.super_kind_gid)
            return res

        def fset(self, value):
            session = object_session(self)
            session.flush()
            oldValue = [i[0] for i in fget(self)]
            oldValue.sort()
            if value:
                self.super_kind_gid = value[0]
            else:
                self.super_kind_gid = None
            newValue = list(value)
            newValue.sort()
            toRemove, toAdd = compareSorted(oldValue, newValue)
            if toRemove:
                session.execute( delete(info_kind_parents).where(and_(info_kind_parents.c.gid == self.gid, authors_users.c.user_id.in_(toRemove))) )
            if toAdd:
                session.execute( insert(info_kind_parents).values([info_kind_parent(self.gid, gid) for gid in toAdd]) )
            session.commit()

        return locals()
    super_gids = property(**super_gids())

    def toDict(self, addExtraArgs = True, inlineExtraArgs = True , selfGid = True, subGids = True, precalculatedGid = False):
        res = {
            "description": self.description,
            "name": self.name,
            "super_names": self.super_names,
        }
        kString = self.kindStr
        if kString != "DocumentContentType":
            res["kindStr"] = kString
        if selfGid:
            res["gid"] = self.gid
        if subGids:
            res["super_gids"] = self.super_gids
        if self.units is not None:
            res["units"] = self.units
        if self.dtypeStr is not None:
            res["dtypeStr"] = self.dtypeStr
        if self.repeats is not None:
            res["repeats"] = self.repeats
        if self.shape is not None:
            res["shape"] = self.shape
        if addExtraArgs:
            if inlineExtraArgs:
                extraArgs = self.extra_args
                if extraArgs:
                    res.update(extraArgs)
            else:
                res["extra_args"] = self.extra_args
        return res

    def has_super_gid(self, gid):
        if self.super_kind_gid == gid:
            return True
        return object_session(self).query(info_kind_parents.c.parent_gid).filter(and_(info_kind_parents.c.info_kind_gid == self.gid, info_kind_parents.c.parent_gid == gid)).first() is not None

    def super_names():
        doc = "The super_names read only property."
        def fget(self):
            vals = object_session(self).query(InfoKind.gid, InfoKind.name).filter(InfoKind.gid == info_kind_parents.c.parent_gid).filter(info_kind_parents.c.info_kind_gid == self.gid).order_by(InfoKind.gid)
            res = []
            firstGid = self.super_kind_gid
            for val in vals:
                if firstGid == val.gid:
                    res.insert(0, val.name)
                else:
                    res.append(val.name)
            return res
        return locals()
    super_names = property(**super_names())

    def __init__(self, session = None, super_gids = None, super_names = None, kindStr = UnsetValue, **kwds):
        Base.__init__(self, **kwds)
        if session:
            session.add(self)
            session.flush()
        if super_gids:
            assert session, "session required for super_gids"
            self.super_gids = super_gids
        if kindStr is UnsetValue:
            if self.kind_gid == None:
                self.kind_gid = metaMetaInfo.gidOf('DocumentContentType')
        else:
            if self.kind_gid is None:
                self.kindStr=kindStr
            assert self.kindStr == kindStr, "kindStr is %s and but stored kind has name %s" % (kindStr, self.kindStr)
        if super_names is not None:
            if super_gids is None:
                if super_names:
                    raise Exception("super_gids required with super_names %s for %s" % (super_names, self.toDict()))
                else:
                    return
            if len(super_names) != len(super_gids):
                raise Exception("incompatible super_names and super_gids")
            # object_session(self).commit()
            sn=self.super_names
            if super_names:
                if super_names[0] != sn[0]:
                    raise Exception("first super type has different name (%r vs %r)" % (sn[0], super_names[0]))
                sn.sort()
                sn2 = list(super_names)
                sn2.sort()
                sn.sort()
                if sn != sn2:
                    raise Exception("unexpected type names in super_names (%r vs %r)" % (sn2, sn))

    _addedBasicTypes=False
    @classmethod
    def ensureBasicTypes(cls, session, force=False):
        cls._addedBasicTypes=True
        for metaInfo in allMetaInfo.values():
            cls.ensureInfoKindEnv(metaInfo, session)

    def kindStr():
        doc = "A string representing the kind of this InfoKind."
        def fget(self):
            if self.kind_gid is None:
                return "MetaType"
            else:
                ik = object_session(self).query(InfoKind.name).filter(InfoKind.gid == self.kind_gid).first()
                if ik:
                    return ik.name
                else:
                    return None

        def fset(self, value):
            if value == "MetaType":
                self.kind_gid = None
            else:
                if value in metaMetaInfo:
                    sha = metaMetaInfo.gidOf(value)
                    self.kind_gid = sha
                else:
                    knownIks = object_session(self).session.query(InfoKind).filter(InfoKind.name == value).filter(InfoKind.kind_gid == None).all()
                    if knownIks:
                        if len(knownIks)>1:
                            logging.getLogger("nomad.InfoKind").warn(
                                    "multiple %s InfoKinds in db detected %s when trying to add %s",
                                    ikEl.kindStr, [x.toInfoKindEl().toDict() for x in knownIks],ikEl)
                        knownIks.sort()
                        self.kind_gid = knownIks[0].gid
                    else:
                        logging.getLogger('nomad.InfoKind').warn('Unknown kind type %s cound not be resolved, replacing with an UnknownType',
                                ikEl.kindStr)
                        kindType = session.query(InfoKind).filter(and_(InfoKind.name == value, InfoKind.kind_gid == metaMetaInfo.gidOf('UnknownType'))).first()
                        if not kindType:
                            logging.getLogger('nomad.InfoKind').warn("will add unknown type for %s", ikEl.kindStr)
                            goInteractive(locals())
                            tmpEnv = InfoKindEnv()
                            tmpEnv.addInfoKindEl(metaMetaInfo.InfoKind('UnknownType'))
                            unkType = InfoKindEl(name = ikEl.kindStr, kindStr = 'UnknownType', description = 'auto inserted unknown type')
                            tmpEnv.addInfoKindEl(unkType)
                            self.kind = InfoKind(session = session, **unkType.toDict(env, inlineExtraArgs=False))
        return locals()
    kindStr = property(**kindStr())

    def egid():
            doc = "The egid property."
            def fget(self):
                    return "K" + self.gid
            def fset(self, value):
                    if len(value)!= 58 and len(value)!= 34:
                            raise Exception("Invalid length for egid value for InfoKind: {0!r}".format(value))
                    if value[0]!="K":
                        raise Exception("Invalid egid value for InfoKind: {0!r}".format(value))
                    self.gid = value[:gidSize]
            return locals()
    egid = property(**egid())

    @staticmethod
    def fromInfoKindEl(ikEl, env, session):
        gid = env.gidOf(ikEl.name)
        existingValue = session.query(InfoKind).get(gid)
        if existingValue is not None:
            return existingValue, False
        kindType = None
        if (ikEl.kindStr == '' or ikEl.kindStr == 'MetaType'):
            if not ikEl.name in metaMetaInfo:
                # ['DocumentType', 'DocumentContentType', 'AbstractDocumentContentType', 'ConnectionType', 'UnknownType']
                # allow???
                raise Exception('New root meta types not allowed, got {0}'.format(ikEl.toDict(env)))
        else:
            # kindStr should uniquely identify the type here we already handle the case that new root types are allowed
            basicKindTypeEl = metaMetaInfo.infoKindEl(ikEl.kindStr)
            envKindTypeEl = env.infoKindEl(ikEl.kindStr)
            if envKindTypeEl and not envKindTypeEl.kindStr in ["", "MetaType"]:
                envKindTypeEl = None
            if basicKindTypeEl and not basicKindTypeEl.kindStr in ["", "MetaType"]: # not a valid meta type, ignore
                basicKindTypeEl = None
            if basicKindTypeEl:
                kindType = session.query(InfoKind).get(metaMetaInfo.gidOf(basicKindTypeEl.name))
                if not kindType:
                    kindType = InfoKind(session = session, kind_gid = metaMetaInfo.gidOf(kindType.kindStr) if not kindType.kindStr in ["", "MetaType"] else None,
                            **basicKindTypeEl.toDict(metaMetaInfo, inlineExtraArgs = False))
                    session.commit()
            if envKindTypeEl is not None:
                if basicKindTypeEl:
                    # defined both in the env and as basic Type
                    if set(envKindTypeEl.super_names) != set(basicKindTypeEl.super_names):
                        logging.getLogger('nomad.InfoKind').warn('incompatible basic type %s vs %s, ignoring new type',
                            ikEl.toDict(), res.toDict())
                    elif gid != metaMetaInfo.gidOf(ikEl.kindStr):
                        logging.getLogger('nomad.InfoKind').warn('ignoring new kind type %s that has different gid but same name as a basic type %s',
                            kindType.toDict(), ikEl.toDict(env))
                else:
                    envKindType = session.query(InfoKind).get(env.gidOf(ikEl.kindStr))
                    if envKindType:
                        kindType = envKindType
                    else:
                        kindType = InfoKind(session = session, **envKindTypeEl.toDict(env, inlineExtraArgs = False))
            if not kindType:
                logging.getLogger('nomad.InfoKind').warn('Unknown kind type %s cound not be resolved, replacing with an UnknownType',
                            ikEl.kindStr)
                kindType = session.query(InfoKind).filter(InfoKind.name == ikEl.kindStr).filter(InfoKind.kind_gid == metaMetaInfo.gidOf('UnknownType')).first()
                if kindType is None:
                    print("will add unknown type for ", ikEl.kindStr)
                    goInteractive(locals())
                    tmpEnv = InfoKindEnv()
                    tmpEnv.addInfoKindEl(metaMetaInfo.infoKindEl('UnknownType'))
                    unkType = InfoKindEl(name = ikEl.kindStr, kindStr = 'UnknownType', description = 'auto inserted unknown type')
                    tmpEnv.addInfoKindEl(unkType)
                    assert metaMetaInfo.gidOf('UnknownType') == tmpEnv.gidOf('UnknownType')
                    kindType = InfoKind(session = session, kind_gid = tmpEnv.gidOf('UnknownType'), **unkType.toDict(tmpEnv, inlineExtraArgs=False))
                    session.commit()
        if ikEl.kindStr == 'DocumentType' or ikEl.kindStr == 'ConnectionType':
            # we would like the name to be enough to identify these, so we warn
            kindIK = aliased(InfoKind)
            dbIks = session.query(InfoKind).filter(InfoKind.name == ikEl.name).join(kindIK, InfoKind.kind).filter(kindIK.name == ikEl.kindStr).all()
            if len(dbIks)>0:
                knownIks=[]
                for dbIk in dbIks:
                    if not metaMetaInfo.gidOf('UnknownDocumentType') in dbIk.cumulative_super_kinds:
                        knownIks.append(dbIk)
                if knownIks:
                    knownIks.sort()
                    if len(knownIks)>1:
                        logging.getLogger("nomad.InfoKind").warn(
                                "multiple %s InfoKinds in db detected %s when trying to add %s",
                                ikEl.kindStr, [x.toInfoKindEl().toDict() for x in knownIks],ikEl)
                    res = knownIks[0]
                    superIK = aliased(InfoKind)
                    superNames = set(session.query(superIK.name).join(superIK, InfoKind.super_kinds).filter(InfoKind.gid == res.gid).all())
                    if superNames != set(ikEl.super_names):
                        logging.getLogger('nomad.InfoKind').warn('incompatible basic type %s vs %s',
                                ikEl.toDict(env), res.toDict())
                    else:
                        logging.getLogger('nomad.InfoKind').warn('found InfoKind %s when trying to add %s',
                                res.toDict(), ikEl.toDict(env))
        ikElDict = ikEl.toDict(env, inlineExtraArgs = False, selfGid = True, subGids = True)
        return InfoKind(session = session, kind = kindType, **ikElDict), True

    def toInfoKindEl(self):
        superKinds=self.super_kinds.all().order_by('name').values_list("name", flat=True)
        superKinds.sort()
        if self.super_kind:
            superKinds.remove(self.super_kind.name)
            superKinds.insert(0,self.super_kind.name)
        return InfoKindEl(name=self.name, description=self.description, kindStr=self.kindStr,
            units=self.units,super_names=superKinds)

    @classmethod
    def ensureInfoKindEnv(cls, env, session): 
        iks = env.sortedIKs()
        added = 0
        existing = 0
        for ik in iks:
            ki, wasAdded = cls.fromInfoKindEl(ik, env, session)
            if wasAdded:
                added += 1
            else:
                existing += 1
        return { 'added': added, 'existing': existing }

    def update_cumulative_super_kinds(self):
        newKinds = set()
        evaluatedKinds = set()
        toAdd=set([self])
        while len(toAdd)>0:
            k = toAdd.pop()
            if not k.publicOverride is None:
                k = k.publicOverride
            if k in evaluatedKinds:
                continue
            evaluatedKinds.add(k)
            newKinds.add(k)
            for k2 in k.super_kinds.all():
                if not k2.publicOverride is None:
                    k2 = k2.publicOverride
                if not k2 in evaluatedKinds:
                    toAdd.add(k2)
        toDel=set()
        for k in self.cumulative_super_kinds.all():
            if not k.publicOverride is None:
                k = k.publicOverride
            if k in newKinds:
                newKinds.remove(k)
            else:
                toDel.add(k)
        if len(toDel) == 0 and len(newKinds) == 0:
            return False
        for k in toDel:
            self.cumulative_super_kinds.remove(k)
        for k in newKinds:
            self.cumulative_super_kinds.add(k)
        return True

    def save(self, *argv, **kwds):
        if self.super_kind and not self.super_kind in self.super_kinds.all():
            self.super_kinds.add(self.super_kind)
        self.update_cumulative_super_kinds()
        super(InfoKind, self).save(*argv, **kwds)

    def __unicode__(self):
        return self.name

    def addPublicOverride(self, ik):
        if self.publicOverride==ik:
            return
        oldV=self.publicOverride
        self.publicOverride=ik
        self.save()
        for info in self.infos:
            info.update_has_info_of_kind()

    @staticmethod
    def connectionEgid():
        return "connectionTypeSK"

    @staticmethod
    def overridesConnectionEgid():
        return "overridesConnectionSK"

    @staticmethod
    def typeEgid():
        return "typeInfoKindSK"

    @staticmethod
    def propertyEgid():
        return "propertyTypeSK"

    def __cmp__(ik1, ik2):
        """orders first non overridden public, then overridden public, then by gid"""
        c = cmp(ik2.public, ik1.public)
        if c!= 0: return c
        c = cmp(ik2.public_override_gid is None, ik1.public_override_gid is None)
        if c!= 0: return c
        c = cmp(ik1.gid, ik2.gid)
        return c



import os

from typing import Tuple, List, Union
import mongomock
import pymongo.collection
from fastapi import HTTPException

from optimade.filterparser import LarkParser
from optimade.filtertransformers.mongo import MongoTransformer
from optimade.server.config import CONFIG
from optimade.models import EntryResource
from optimade.server.mappers import BaseResourceMapper
from optimade.server.query_params import EntryListingQueryParams, SingleEntryQueryParams
from .entry_collections import EntryCollection

try:
    CI_FORCE_MONGO = bool(int(os.environ.get("OPTIMADE_CI_FORCE_MONGO", 0)))
except (TypeError, ValueError):  # pragma: no cover
    CI_FORCE_MONGO = False


if CONFIG.use_real_mongo or CI_FORCE_MONGO:
    from pymongo import MongoClient

    client = MongoClient(CONFIG.mongo_uri)
    print("Using: Real MongoDB (pymongo)")
else:
    from mongomock import MongoClient

    client = MongoClient()
    print("Using: Mock MongoDB (mongomock)")


class MongoCollection(EntryCollection):
    def __init__(
        self,
        collection: Union[
            pymongo.collection.Collection, mongomock.collection.Collection
        ],
        resource_cls: EntryResource,
        resource_mapper: BaseResourceMapper,
    ):
        super().__init__(collection, resource_cls, resource_mapper)
        self.transformer = MongoTransformer(mapper=resource_mapper)

        self.provider_prefix = CONFIG.provider.prefix
        self.provider_fields = CONFIG.provider_fields.get(resource_mapper.ENDPOINT, [])
        self.parser = LarkParser(
            version=(0, 10, 1), variant="default"
        )  # The MongoTransformer only supports v0.10.1 as the latest grammar

        # check aliases do not clash with mongo operators
        self._check_aliases(self.resource_mapper.all_aliases())
        self._check_aliases(self.resource_mapper.all_length_aliases())

    def __len__(self):
        return self.collection.estimated_document_count()

    def __contains__(self, entry):
        return self.collection.count_documents(entry.dict()) > 0

    def count(self, **kwargs):
        for k in list(kwargs.keys()):
            if k not in ("filter", "skip", "limit", "hint", "maxTimeMS"):
                del kwargs[k]
        if "filter" not in kwargs:  # "filter" is needed for count_documents()
            kwargs["filter"] = {}
        return self.collection.count_documents(**kwargs)

    def find(
        self, params: Union[EntryListingQueryParams, SingleEntryQueryParams]
    ) -> Tuple[List[EntryResource], int, bool, set]:
        criteria = self._parse_params(params)

        all_fields = criteria.pop("fields")
        if getattr(params, "response_fields", False):
            fields = set(params.response_fields.split(","))
            fields |= self.resource_mapper.get_required_fields()
        else:
            fields = all_fields.copy()

        results = []
        for doc in self.collection.find(**criteria):
            results.append(self.resource_cls(**self.resource_mapper.map_back(doc)))

        nresults_now = len(results)
        if isinstance(params, EntryListingQueryParams):
            criteria_nolimit = criteria.copy()
            criteria_nolimit.pop("limit", None)
            data_returned = self.count(**criteria_nolimit)
            more_data_available = nresults_now < data_returned
        else:
            # SingleEntryQueryParams, e.g., /structures/{entry_id}
            data_returned = nresults_now
            more_data_available = False
            if nresults_now > 1:
                raise HTTPException(
                    status_code=404,
                    detail=f"Instead of a single entry, {nresults_now} entries were found",
                )
            results = results[0] if results else None

        return results, data_returned, more_data_available, all_fields - fields

    def _parse_params(
        self, params: Union[EntryListingQueryParams, SingleEntryQueryParams]
    ) -> dict:
        cursor_kwargs = {}

        if getattr(params, "filter", False):
            tree = self.parser.parse(params.filter)
            cursor_kwargs["filter"] = self.transformer.transform(tree)
        else:
            cursor_kwargs["filter"] = {}

        if (
            getattr(params, "response_format", False)
            and params.response_format != "json"
        ):
            raise HTTPException(
                status_code=400, detail="Only 'json' response_format supported"
            )

        if getattr(params, "page_limit", False):
            limit = params.page_limit
            if limit > CONFIG.page_limit_max:
                raise HTTPException(
                    status_code=403,  # Forbidden
                    detail=f"Max allowed page_limit is {CONFIG.page_limit_max}, you requested {limit}",
                )
            cursor_kwargs["limit"] = limit
        else:
            cursor_kwargs["limit"] = CONFIG.page_limit

        # All OPTIMADE fields
        fields = self.resource_mapper.TOP_LEVEL_NON_ATTRIBUTES_FIELDS.copy()
        fields |= self.get_attribute_fields()
        # All provider-specific fields
        fields |= {
            f"_{self.provider_prefix}_{field_name}"
            for field_name in self.provider_fields
        }
        cursor_kwargs["fields"] = fields
        cursor_kwargs["projection"] = [
            self.resource_mapper.alias_for(f) for f in fields
        ]

        if getattr(params, "sort", False):
            sort_spec = []
            for elt in params.sort.split(","):
                field = elt
                sort_dir = 1
                if elt.startswith("-"):
                    field = field[1:]
                    sort_dir = -1
                sort_spec.append((field, sort_dir))
            cursor_kwargs["sort"] = sort_spec

        if getattr(params, "page_offset", False):
            cursor_kwargs["skip"] = params.page_offset

        return cursor_kwargs

    def _check_aliases(self, aliases):
        """ Check that aliases do not clash with mongo keywords. """
        if any(
            alias[0].startswith("$") or alias[1].startswith("$") for alias in aliases
        ):
            raise RuntimeError(f"Cannot define an alias starting with a '$': {aliases}")



from builtins import object
from nomadcore.simple_parser import mainFunction, CachingLevel
from nomadcore.simple_parser import SimpleMatcher as SM
from nomadcore.local_meta_info import loadJsonFile, InfoKindEl
import os, sys, json, logging
import numpy as np


################################################################
# This is the subparser for the main fleur input file (.inp)
################################################################

# Copyright 2016-2018 Daria M. Tomecka, Fawzi Mohamed
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = "Daria M. Tomecka"
__maintainer__ = "Daria M. Tomecka"
__email__ = "tomeckadm@gmail.com;"
__date__ = "15/05/2017"

class FleurInpContext(object):
    """context for wien2k struct parser"""

    def __init__(self):
        self.parser = None

    def initialize_values(self):
        """allows to reset values if the same superContext is used to parse different files"""
        pass

    def startedParsing(self, path, parser):
        """called when parsing starts"""
        self.parser = parser
        # allows to reset values if the same superContext is used to parse different files
        self.initialize_values()

    def onClose_section_system(self, backend, gIndex, section):
        equiv_atoms = section["x_fleur_section_equiv_atoms"]
        #logging.info("section: %s", section)
        labels = []
        pos = []
        for eqAtoms in equiv_atoms:
            label = eqAtoms["x_fleur_atom_name"][0]
            x = eqAtoms["x_fleur_atom_pos_x"]
            y = eqAtoms["x_fleur_atom_pos_y"]
            z = eqAtoms["x_fleur_atom_pos_z"]
            if len(x) != len(y) or len(x) != len(z):
                raise Exception("incorrect parsing, different number of x,y,z components")
            groupPos = [[x[i],y[i],z[i]] for i in range(len(x))]
            nAt = len(groupPos)
            labels += [label for i in range(nAt)]
            pos += groupPos
        backend.addValue("atom_labels", labels)
        backend.addValue("atom_positions", pos)


# description of the input
def buildStructureMatchers():
    return SM(
    name = 'root',
    weak = True,
    startReStr = "",
    sections = ["section_run", "section_system"],
    subMatchers = [
        SM(name = 'systemName',
           startReStr = r"\s*strho.*\n(?P<x_fleur_system_namein>.*)"),#L2
        SM(r"\s{2}(?P<<x_fleur_nr_of_atom_types>[0-9]+)\n",#L10
#
           sections=["x_fleur_section_equiv_atoms"],
           subMatchers=[
               SM(r"\W{35}\n*(?P<x_fleur_name_of_atom_type>\w*)\s*(?P<x_fleur_nuclear_number>[0-9].*)\s(?P<x_fleurnumber_of_core_levels>[0-9].*)\s*(?P<l-expansion_cutoff>[0-9].)\s(?P<x_fleur_mt_gridpoints>[0-9].*)\s(?P<x_fleur_mt_radius>[0-9].+)\s(?P<x_fleur_logarythmic_increment>[0-9].*)"),#l11
               SM(r"(?P<number_equiv_atoms_in_this_atom_type>[0-9].*),force\s=.*"),# number of equivalent atoms in this atom type
               SM(r"\s*force.*\n\s*(?P<x_fleur_atom_pos_x>[-+0-9.]+)\s+(?P<x_fleur_atom_pos_y>[-+0-9.]+)\s+(?P<x_fleur_atom_pos_z>[-+0-9.]+)\s+(?P<x_fleur_atom_coord_scale>[-+0-9.]+)\n",
                  repeats=True,
                  )
                        ],
               repeat = True
        # SM(r"\s\s(?P<add>[0-9.]*)\s\s(?P<addd>[0-9.]*)\nvchk.*"),
       )
    ])

def get_cachingLevelForMetaName(metaInfoEnv, CachingLvl):
    """Sets the caching level for the metadata.

    Args:
        metaInfoEnv: metadata which is an object of the class InfoKindEnv in nomadcore.local_meta_info.py.
        CachingLvl: Sets the CachingLevel for the sections k_band, run, and single_configuration_calculation.
            This allows to run the parser without opening new sections.

    Returns:
        Dictionary with metaname as key and caching level as value.
    """
    # manually adjust caching of metadata
    cachingLevelForMetaName = {
                               'section_run': CachingLvl,
                               'section_system': CachingLvl
                              }
    cachingLevelForMetaName["x_fleur_system_namein"] = CachingLevel.ForwardAndCache
    cachingLevelForMetaName["x_fleur_section_equiv_atoms"] = CachingLevel.ForwardAndCache
    return cachingLevelForMetaName

# loading metadata from nomad-meta-info/meta_info/nomad_meta_info/fhi_aims.nomadmetainfo.json



# Copyright 2018 Fawzi Mohamed, Danio Brambila, Markus Scheidgen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an"AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os.path
import json
import numpy as np

from nomad.normalizing.normalizer import Normalizer

controlIn_basis_set = 'x_fhi_aims_section_controlIn_basis_set'
controlIn_basis_func = 'x_fhi_aims_section_controlIn_basis_func'
controlIn_nucleus = 'x_fhi_aims_controlIn_nucleus'

pure_types_json = dict()

for pure_types_str in ['light', 'really_tight', 'tight']:
    with open(os.path.join(os.path.dirname(__file__), 'data', pure_types_str + '.json')) as f:
        json_data = json.load(f)
        section_method = json_data['sections']['section_run-0']['sections']['section_method-0']
        pure_types_json[pure_types_str] = section_method[controlIn_basis_set]


class FhiAimsBaseNormalizer(Normalizer):

    # Finds out if val is in the array
    def compare_val_list(self, val, list):
        if val in list:
            return 0
        return 1

    # Comparison between two dicts.
    # List structure:
    def compare_dict_dict(self, d1, d2):
        sum2 = np.zeros(len(d2))

        # Loop over the size of dict2
        for k in np.arange(0, len(d2)):
            # Lopp over the elements of each dict2
            for idx, val in d1.items():
                # Excludes the keys that are always different.
                if (idx not in ["gIndex", "references", "uri"]):
                    try:
                        if (val != d2[k][idx]):
                            sum2[k] = sum2[k] + 1
                    except KeyError:  # this exception case arises if the cut off potential is not a number
                        continue
        if (min(sum2) == 0):
            return 0
        else:
            return 1  # sum(sum2)

    def compare_to_defaults(self, dict2_default, dict1):
        # first compare the integration grid
        false_hits_integration_grid = 0
        false_hits_basis = 0

        for key in dict1:
            if key not in ['gIndex', 'uri', controlIn_basis_func]:
                if np.size(dict1[key]) == 1:
                    if(dict1[key] != dict2_default[key]):
                        false_hits_integration_grid += 1
                        false_hits_integration_grid += abs(np.size(dict1[key]) - np.size(dict2_default[key]))
                if np.size(dict1[key]) > 1:
                    for i in dict1[key]:
                        false_hits_integration_grid += self.compare_val_list(i, dict2_default[key])
                        false_hits_integration_grid += abs(np.size(dict1[key]) - np.size(dict2_default[key]))

            elif (key == controlIn_basis_func):
                for i in np.arange(0, len(dict1[key])):
                    false_hits_basis += self.compare_dict_dict(
                        dict1[key][i], dict2_default[key])
                false_hits_basis += abs(len(dict1[key]) - len(dict2_default[key]))

        return [false_hits_integration_grid, false_hits_basis]

    def normalize(self, logger=None) -> None:
        super().normalize(logger)
        if not self.section_run or self.section_run.program_name != 'FHI-aims':
            return

        for method in self.section_run.section_method:
            to_compare = getattr(method, controlIn_basis_set, None)
            if to_compare is None:
                # not fhi aims data
                continue

            matrix_hits_int = dict.fromkeys(pure_types_json, 0)
            matrix_hits_basis = dict.fromkeys(pure_types_json, 0)

            for index, data in enumerate(to_compare):
                atom_index = int(data[controlIn_nucleus])
                for key, val in pure_types_json.items():
                    results = self.compare_to_defaults(val[atom_index], to_compare[index])
                    matrix_hits_int[key] += results[0]
                    matrix_hits_basis[key] += results[1]

                    # matrix_hits[key]=matrix_hits[key]+CompareToDefaults(val[AtomIndex],to_compare[i])

            closest_base_int = min(matrix_hits_int, key=matrix_hits_int.get)
            if (matrix_hits_basis[min(matrix_hits_basis, key=matrix_hits_basis.get)] == 0):
                closest_base_base = ''
            else:
                closest_base_base = '+'

            if (matrix_hits_int[closest_base_int] == 0):
                method.basis_set = closest_base_int + closest_base_base
            elif(matrix_hits_int[closest_base_int] <= 5):
                method.basis_set = '~' + closest_base_int + closest_base_base
            elif(matrix_hits_int[closest_base_int] > 5):
                method.basis_set = 'custom-' + closest_base_int


# import setup_paths
# import json
# from  numpy import *
# import os.path, glob
# from nomadcore.local_meta_info import loadJsonFile, InfoKindEl
# from nomadcore.parser_backend import JsonParseEventsWriterBackend
# from nomadcore.parse_streamed_dicts import *
# import sys

# import logging

# base_path = os.path.abspath(os.path.dirname(__file__))

# files = glob.glob(os.path.join(base_path,"json_default/*.json"))

# # Open json files
# def open_json(path):
#   default=dict()
#   with open(path) as data_file:
#     default = json.load(data_file)
#   return default

# # Finds out if val is in the array
# def CompareList(val,list):
#   if val in list:
#     return 0
#   return 1

# # Comparison between two dicts.
# # List structure:
# def CompareTwoDicts(d1,d2):
#   sum2=zeros(len(d2))

# # Loop over the size of dict2
#   for k in arange(0,len(d2)):
# #   Lopp over the elements of each dict2
#     for idx,val in d1.items():
#       if (idx not in ["gIndex","references", "uri"]) : # Excludes the keys that are always different.
#         try:
#           if (val!=d2[k][idx]):
#             sum2[k]=sum2[k]+1
#         except KeyError: # this exception case arises if the cut off potential is not a number
#           continue
#   if (min(sum2)==0):
#     return 0
#   else:
#     return 1# sum(sum2)

# def CompareToDefaults(dict2_default,dict1):

# #first compare the integration grid
#   false_hits_integration_grid=0
#   false_hits_basis=0
#   for key in dict1:
#     if key not in ["gIndex", "uri", "x_fhi_aims_section_controlIn_basis_func"]:
#       if size(dict1[key])==1:
#         if(dict1[key]!=dict2_default[key]):
#           false_hits_integration_grid+=1
#         false_hits_integration_grid+=abs(size(dict1[key])-size(dict2_default[key]))
#       if size(dict1[key])>1:
#         for i in dict1[key]:
#            false_hits_integration_grid+=CompareList(i,dict2_default[key])
#         false_hits_integration_grid+=abs(size(dict1[key])-size(dict2_default[key]))
#     elif (key == "x_fhi_aims_section_controlIn_basis_func"):
#       for i in arange(0,len(dict1[key])):
#           false_hits_basis+=CompareTwoDicts(dict1[key][i],dict2_default[key])
#       false_hits_basis+=abs(len(dict1[key])-len(dict2_default[key]))
#   return [false_hits_integration_grid,false_hits_basis]

# import sys, getopt
# def main():
#   metapath = '../../../../nomad-meta-info/meta_info/nomad_meta_info/' +\
#         'common.nomadmetainfo.json'
#   metaInfoPath = os.path.normpath(
#     os.path.join(os.path.dirname(os.path.abspath(__file__)), metapath))

#   metaInfoEnv, warns = loadJsonFile(filePath=metaInfoPath,
#                                     dependencyLoader=None,
#                                     extraArgsHandling=InfoKindEl.ADD_EXTRA_ARGS,
#                                     uri=None)
#   backend = JsonParseEventsWriterBackend(metaInfoEnv)
#   calcContext = sys.argv[1]
#   backend.startedParsingSession(
#     calcContext,
#     parserInfo = {'name':'FhiAimsBasisNormalizer', 'version': '1.0'})

#   pure_types_json={}
#   matrix_hits_int={}
#   matrix_hits_basis={}


#   A=ParseStreamedDicts(sys.stdin)

#   while True:
#     to_compare=A.readNextDict()
#     if to_compare is None:
#       break
#     context=to_compare["context"]  #uncomment for prod
#     try:
#       to_compare=to_compare["section_method"]["x_fhi_aims_section_controlIn_basis_set"]
#     except:
#       continue

#     matrix_hits_int = dict.fromkeys(matrix_hits_int, 0)
#     matrix_hits_basis = dict.fromkeys(matrix_hits_basis, 0)

#     for i,d in enumerate(to_compare):
#        AtomIndex=int(d["x_fhi_aims_controlIn_nucleus"])
#        for key,val in pure_types_json.items():
#           res=CompareToDefaults(val[AtomIndex],to_compare[i])
#           matrix_hits_int[key]+=res[0]
#           matrix_hits_basis[key]+=res[1]

# #          matrix_hits[key]=matrix_hits[key]+CompareToDefaults(val[AtomIndex],to_compare[i])


#     closest_base_int=min(matrix_hits_int, key=matrix_hits_int.get)
#     if (matrix_hits_basis[min(matrix_hits_basis, key=matrix_hits_basis.get)] ==0):
#        closest_base_base=''
#     else:
#        closest_base_base='+'
#     if (matrix_hits_int[closest_base_int]==0):
# #      print(closest_base_int +closest_base_base)
#       backend.addValue("basis_set",closest_base_int +closest_base_base)
#     elif(matrix_hits_int[closest_base_int]<=5):
# #      print('~'+closest_base_int+closest_base_base)
#       backend.addValue("basis_set","~"+closest_base_int +closest_base_base)
#     elif(matrix_hits_int[closest_base_int]>5):
#       backend.addValue("basis_set",'custom-'+closest_base_int)
# #      print('custom-'+closest_base_int)

#   backend.finishedParsingSession("ParseSuccess", None)
#   sys.stdout.flush()
#   return

# if __name__ == "__main__":
#     main()



# Copyright 2015-2018 Martina Stella, Massimo Riello, Fawzi Mohamed, Ankit Kariryaa
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from builtins import range
from builtins import object
import numpy as np
import nomadcore.ActivateLogging
from nomadcore.caching_backend import CachingLevel
from nomadcore.simple_parser import mainFunction
from nomadcore.simple_parser import SimpleMatcher as SM
from nomadcore.local_meta_info import loadJsonFile, InfoKindEl
from castepparser.CastepCommon import get_metaInfo
import logging, os, re, sys



############################################################
# This is the parser for the *.cell file of CASTEP.
############################################################

logger = logging.getLogger("nomad.CastepCellParser")

class CastepCellParserContext(object):
    """Context for parsing CASTEP *.cell file.


    The onClose_ functions allow processing and writing of cached values after a section is closed.
    They take the following arguments:
        backend: Class that takes care of wrting and caching of metadata.
        gIndex: Index of the section that is closed.
        section: The cached values and sections that were found in the section that is closed.
    """
    def __init__(self, writeMetaData = True):
        """Args:
            writeMetaData: Deteremines if metadata is written or stored in class attributes.
        """
        self.writeMetaData = writeMetaData

    def startedParsing(self, fInName, parser):
        """Function is called when the parsing starts and the compiled parser is obtained.

        Args:
            fInName: The file name on which the current parser is running.
            parser: The compiled parser. Is an object of the class SimpleParser in nomadcore.simple_parser.py.
        """
        self.parser = parser
        # get unit from metadata for band energies
        # allows to reset values if the same superContext is used to parse different files
        self.band_energies = None
        self.band_k_points = None
        self.band_occupations = None
        self.klabel_seg = []
        self.k_crd = []
        self.k_sgt_start_end = []

    def onClose_section_k_band(self, backend, gIndex, section):
        """Trigger called when section_k_band is closed.

        Store the parsed values and write them if writeMetaData is True.
        """
        k_p = section['x_castep_store_k_path']

        k_count = len(k_p)
        self.k_crd = []
        for i in range(0, k_count):
            k_p[i] = k_p[i].split()
            k_p[i] = [float(j) for j in k_p[i]]
            k_p_int = k_p[i]
            self.k_crd.append(k_p_int)


        self.k_sgt_start_end = []
        for i in range(k_count-1):
            k_sgt = [ self.k_crd[i], self.k_crd[i+1] ]
            self.k_sgt_start_end.append(k_sgt)


        backend.addArrayValues('x_castep_k_path', np.asarray(self.k_crd))
        self.klabel_seg = section['x_castep_store_k_label']

        # backend.addArrayValues('band_segm_start_end', np.asarray(self.k_sgt_start_end))
        # backend.addValue('number_of_k_point_segments', len(self.k_sgt_start_end))


def build_CastepCellFileSimpleMatcher():
    """Builds the SimpleMatcher to parse the *.cell file of CASTEP.

    SimpleMatchers are called with 'SM (' as this string has length 4,
    which allows nice formating of nested SimpleMatchers in python.

    Returns:
       SimpleMatcher that parses *.cell file of CASTEP.
    """
    return SM (name = 'Root1',
        startReStr = "",
        sections = ['section_run'],
        forwardMatch = True,
        weak = True,
        subMatchers = [
        SM (name = 'Root2',
            startReStr = "",
            sections = ['section_single_configuration_calculation'],
            forwardMatch = True,
            weak = True,
            subMatchers = [

            SM(startReStr = r"\s*\%block bs\_kpoint\_path\s*",
                  sections = ['section_k_band'],
                  forwardMatch = True,
                  subMatchers = [

                     SM (r"(?P<x_castep_store_k_path>[\d\.]+\s+[\d\.]+\s+[\d\.]+)\s*\!\s(?P<x_castep_store_k_label>[A-Z])", repeats = True)
                                 ]),

            SM(startReStr = r"\s*\%BLOCK BS\_KPOINT\_PATH\s*",
                  sections = ['section_k_band'],
                  forwardMatch = True,
                  subMatchers = [

                      SM (r"(?P<x_castep_store_k_path>[\d\.]+\s+[\d\.]+\s+[\d\.]+)\s*\!\s(?P<x_castep_store_k_label>[A-Z])", repeats = True)

                                 ]),


            # SM(startReStr = r"\s*\%block kpoint\_list\s*",
            #       sections = ['section_k_band'],
            #       forwardMatch = True,
            #       subMatchers = [

            #          SM (r"\s*(?P<x_castep_store_k_path>[\d\.]+\s+[\d\.]+\s+[\d\.]+[\d\.]+)", repeats = True)
            #                      ]),

            # SM(startReStr = r"\s*\%BLOCK KPOINT\_LIST\s*",
            #       sections = ['section_k_band'],
            #       forwardMatch = True,
            #       subMatchers = [

            #           SM (r"\s*(?P<x_castep_store_k_path>[\d\.]+\s+[\d\.]+\s+[\d\.]+[\d\.]+)", repeats = True)

            #                      ]),

            SM(startReStr = r"\%block kpoint\_list\s*",
                  sections = ['section_k_band'],
                  forwardMatch = True,
                  subMatchers = [

                     SM (r"\s*(?P<x_castep_store_k_path>[-\d\.]+\s+[-\d\.]+\s+[-\d\.]+)\s*\!\s(?P<x_castep_store_k_label>[A-Z])", repeats = True)
                                 ]),

            # SM(startReStr = r"\%BLOCK KPOINT\_LIST\s*",
            #       sections = ['section_k_band'],
            #       forwardMatch = True,
            #       subMatchers = [

            #           SM (r"\s*(?P<x_castep_store_k_path>[\d\.]+\s+[\d\.]+\s+[\d\.]+[\d\.]+)", repeats = True)

            #                      ]),
            #SM (name = 'Root3',
            #    startReStr = r"\s*\%block bs\_kpoint\_path\s*",
            #    sections = ['section_k_band'],
            #    forwardMatch = True,
            #    weak = True,
            #    subMatchers = [
            #    SM (r"(?P<castep_store_k_path>[\d\.]+\s+[\d\.]+\s+[\d\.]+)", repeats = True)
            #    ]),

            ])

        ])


def get_cachingLevelForMetaName(metaInfoEnv, CachingLvl):
    """Sets the caching level for the metadata.

    Args:
        metaInfoEnv: metadata which is an object of the class InfoKindEnv in nomadcore.local_meta_info.py.
        CachingLvl: Sets the CachingLevel for the sections k_band, run, and single_configuration_calculation.
            This allows to run the parser without opening new sections.

    Returns:
        Dictionary with metaname as key and caching level as value.
    """
    # manually adjust caching of metadata
    cachingLevelForMetaName = {
                               'section_k_band': CachingLvl,
                               'section_run': CachingLvl,
                               'section_single_configuration_calculation': CachingLvl,
                              }
    # Set all band metadata to Cache as they need post-processsing.
    for name in metaInfoEnv.infoKinds:
        if name.startswith('x_castep_'):
            cachingLevelForMetaName[name] = CachingLevel.Cache
    return cachingLevelForMetaName


def main(CachingLvl):
    """Main function.

    Set up everything for the parsing of the CASTEP *.cell file and run the parsing.

    Args:
        CachingLvl: Sets the CachingLevel for the sections k_band, run, and single_configuration_calculation.
            This allows to run the parser without opening new sections.
    """
    # get band.out file description
    CasteCellFileSimpleMatcher = build_CastepCellFileSimpleMatcher()
    # loading metadata from nomad-meta-info/meta_info/nomad_meta_info/castep.nomadmetainfo.json
    metaInfoPath = os.path.normpath(os.path.join(os.path.dirname(os.path.abspath(__file__)),"../../../../nomad-meta-info/meta_info/nomad_meta_info/castep.nomadmetainfo.json"))
    metaInfoEnv = get_metaInfo(metaInfoPath)
    # set parser info
    parserInfo = {'name':'castep-cell-parser', 'version': '1.0'}
    # get caching level for metadata
    cachingLevelForMetaName = get_cachingLevelForMetaName(metaInfoEnv, CachingLvl)
    # start parsing
    mainFunction(mainFileDescription = CasteCellFileSimpleMatcher,
                 metaInfoEnv = metaInfoEnv,
                 parserInfo = parserInfo,
                 cachingLevelForMetaName = cachingLevelForMetaName,
                 superContext = CastepCellParserContext())

if __name__ == "__main__":
    main(CachingLevel.Forward)










# Copyright 2016-2018 Arvid Conrad Ihrig, Aliaksei Mazheika
#                     Fritz-Haber-Institut der Max-Planck-Gesellschaft
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module constructs the parser for the ESCF module from TurboMole"""

import logging
import re
from nomadcore.simple_parser import SimpleMatcher as SM
import turbomoleparser.TurbomoleCommon as Common

logger = logging.getLogger("nomad.turbomoleParser")

class ESCFparser(object):

    def __init__(self, context, key="escf"):
        context[key] = self
        self.__context = context
        self.__backend = None

    def purge_data(self):
        pass

    def set_backend(self, backend):
        self.__backend = backend

    def build_parser(self):
        references = SM(r"\s{5,}[^+ ]+",
                        name="references",
                        coverageIgnore=True,
                        repeats=True,
                        )
        header = SM(r"\s*e s c f\s*$",
                    name="Credits",
                    coverageIgnore=True,
                    subMatchers=[references],
                    endReStr=r"\s*\+-+\+"
                    )

        sub_matchers = [
            header,
            self.__context["geo"].build_qm_geometry_matcher(),
            self.__context["geo"].build_orbital_basis_matcher(),
            self.__build_gw_matcher(),
        ]

        return self.__context.build_module_matcher("escf", sub_matchers, "ESCF")

    def __build_gw_matcher(self):
        def get_gw_approximation(backend, groups):
            types = groups[0]
            regex = re.compile(r"\s*(?P<index>[0-9]+)\s*:\s*(?P<name>[^\s:]+)")
            approximations = dict()
            match = regex.match(types)
            while match:
                approximations[match.group("index")] = match.group("name")
                match = regex.match(types, pos=match.end(2))
            backend.addValue("electronic_structure_method", "G0W0")
            backend.addValue("calculation_method_kind", "perturbative")
            backend.addValue("x_turbomole_gw_approximation", approximations[groups[1]])

        params = SM(name="GW parameters",
                    startReStr="\s*par[ae]meters:",  # typo in Turbomole 6.6 output
                    sections=["section_method"],
                    fixedStartValues={"number_of_eigenvalues_kpoints": 1},  # TM has no periodic GW
                    subMatchers=[
                        SM(r"\s*number of levels to calculate\s+(?P<number_of_eigenvalues>[0-9]+)",
                           name="num states"),
                        SM(r"\s*number of spin channels\s+(?P<number_of_spin_channels>[0-9]+)",
                           name="num spin channels"),
                        SM(r"\s*type of gw((?:\s+[0-9]+\s*:\s*\S+)+)\s+([0-9]+)\s*$",
                           name="GW approximation",
                           startReAction=get_gw_approximation),
                        SM(r"\s*rpa response function\s+(?P<x_turbomole_gw_use_rpa_response>[TF])",
                           name="GW screened interaction"),
                        SM(r"\s*eta \(Hartree\)\s+(?P<x_turbomole_gw_eta_factor__hartree>"
                           r"[+-]?[0-9]+.?[0-9]*)",
                           name="GW eta factor")
                    ]
                    )

        return SM(name="GW",
                  startReStr=r"\s*GW version\s+[0-9]+",
                  # startReStr = r"\s*GW version\s+(?P<x_turbomole_version_GW>[0-9]+)",
                  sections=["section_eigenvalues"],
                  subMatchers=[
                      params,
                      self.__build_gw_qp_states_matcher_no_spin()
                  ]
                  )

    def __build_gw_qp_states_matcher_no_spin(self):
        state = SM(r"\s*[0-9]+\s+(?P<x_turbomole_eigenvalue_ks_GroundState__eV>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)\s+"
                   r"(?P<x_turbomole_eigenvalue_quasiParticle_energy__eV>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)\s+"
                   r"(?P<x_turbomole_eigenvalue_ExchangeCorrelation_perturbativeGW__eV>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)\s+"
                   r"(?P<x_turbomole_eigenvalue_ExactExchange_perturbativeGW__eV>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)\s+"
                   r"(?P<x_turbomole_eigenvalue_correlation_perturbativeGW__eV>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)\s+"
                   r"(?P<x_turbomole_eigenvalue_ks_ExchangeCorrelation__eV>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)\s+"
                   r"(?P<x_turbomole_Z_factor>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)\s+"
                   r"(?P<x_turbomole_ExchangeCorrelation_perturbativeGW_derivation>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)",
                   name="GW QP state",
                   repeats=True)
        return SM(r"\s*orb\s+eps\s+QP-eps\s+Sigma\s+Sigma_x\s+Sigma_c\s+Vxc\s+Z\s+dS\/de",
                  name="GW QP statelist",
                  sections=["x_turbomole_section_eigenvalues_GW"],
                  subMatchers=[
                      SM(r"\s*in\s*eV", required=True, name="GW output unit"),
                      SM(r"\s*----*", name="<format>", coverageIgnore=True),
                      state,
                      SM(r"\s*----*", name="<format>", coverageIgnore=True),
                      state.copy(),
                      SM(r"\s*----*", name="<format>", coverageIgnore=True),
                  ])
